from __future__ import annotations
from itertools import chain

from e3.aws.cfn import Stack, Join
from e3.aws.cfn.arch.security import amazon_security_groups, github_security_groups
from e3.aws.cfn.ec2 import (
    EC2NetworkInterface,
    EIP,
    Instance,
    InternetGateway,
    LaunchTemplate,
    NatGateway,
    NetworkInterface,
    Route,
    RouteTable,
    Subnet,
    SubnetRouteTableAssociation,
    VPC,
    VPCEndpoint,
    VPCInterfaceEndpoint,
    VPCGatewayAttachment,
)
from e3.aws.cfn.ec2.security import (
    Ipv4EgressRule,
    Ipv4IngressRule,
    PrefixListEgressRule,
    SecurityGroup,
)
from e3.aws.cfn.iam import PolicyDocument, Principal, PrincipalKind, InstanceRole, Allow
from e3.aws.cfn.s3 import Bucket


# Prefix lists are static name used to select a list of IPs for a given
# AWS services. Currently Amazon only offer prefix lists for s3 and
# and dynamodb
PREFIX_LISTS = {
    "eu-west-1": {"s3": "pl-6da54004", "dynamodb": "pl-6fa54006"},
    "us-east-1": {"s3": "pl-63a5400a", "dynamodb": "pl-02cd2c6b"},
}


class AWSFortressError(Exception):
    """Error raised when Fortress configuration fails."""

    pass


class SubnetStack(Stack):
    """Create a subnet with a route table."""

    def __init__(self, name, vpc, cidr_block, description=None):
        """Initialize a subnet.

        This block create a basic subnet with an empty route table

        :param name: logical name of the subnet in the stack
        :type name: str
        :param vpc: VPC containing the subnet
        :type vpc: VPC
        :param cidr_block: block of addresses associated with the subnet
        :type cidr_block: str
        :param description: optional description
        :type description: str | None
        """
        super().__init__(name, description)

        # Create the subnet
        self.add(Subnet(name, vpc, cidr_block))

        # Associate a route table
        self.add(RouteTable(name + "RouteTable", vpc))
        self.add(
            SubnetRouteTableAssociation(
                name + "RouteTableAssoc", self.subnet, self.route_table
            )
        )

    def add_bucket_access(self, bucket_list):
        """Authorize access to a list of buckets using vpc endpoint.

        Note that this just allow an instance in the vpc to ask access
        to a given bucket through the endpoint. This does not change
        the bucket policy.

        The function creates also automatically the S3 VPC endpoint
        on the first call.

        :param bucket_list: list of bucket names
        :type bucket_list: list[Bucket] | list[str]
        """
        if self.name + "S3EndPoint" not in self:
            self.add(
                VPCEndpoint(
                    self.name + "S3EndPoint",
                    "s3",
                    self.subnet.vpc,
                    [self.route_table],
                    PolicyDocument(),
                )
            )
        for bucket in bucket_list:
            if isinstance(bucket, Bucket):
                bucket_name = bucket.ref
            else:
                bucket_name = bucket
            self.s3_endpoint.policy_document.append(
                Allow(
                    to="s3:*",
                    on=[
                        Join(["arn:aws:s3:::", bucket_name]),
                        Join(["arn:aws:s3:::", bucket_name, "/*"]),
                    ],
                    apply_to=Principal(PrincipalKind.EVERYONE),
                )
            )

    @property
    def s3_endpoint(self):
        return self[self.name + "S3EndPoint"]

    @property
    def subnet(self):
        return self[self.name]

    @property
    def cidr_block(self):
        return self[self.name].cidr_block

    @property
    def route_table(self):
        return self[self.name + "RouteTable"]


class VPCStack(Stack):
    """VPC stack.

    Handle a VPC with various networks elements such as subnets and gateways.
    """

    def __init__(self, name, cidr_block, description=None):
        """Create a VPC stack.

        :param name: stack name
        :type name: str
        :param cidr_block: ipv4 address range for the vpc
        :type cird_block: str
        :param description: optional description
        :type description: str | None
        """
        super().__init__(name, description=description)
        self.add(VPC(self.name, cidr_block))
        self.add(InternetGateway(self.name + "InternetGateway"))
        self.add(VPCGatewayAttachment(self.name + "GateLink", self.vpc, self.gateway))

    @property
    def region(self):
        """Region in which the stack is allocated.

        :return: a region
        :rtype: str
        """
        return self[self.name].region

    def add_subnet(self, name, cidr_block, is_public=False, use_nat=False, nat_to=None):
        """Add a subnet.

        :param name: subnet logical name in the stack
        :type name: str
        :param cidr_block: address range of the subnet. Should be a subnet
            of the vpc address range (no check done).
        :type cidr_block: str
        :param is_public: if True create a public subnet. This means that
            a route is created automatically to the vpc internet gateway.
            (default: False)
        :type is_public: bool
        :param use_nat: if True and is_public is True, then add a NAT
            gateway that can be reused by private subnets.
            (default: False)
        :type use_nat: bool
        :param nat_to: if is_public is False and nat_to is a string,
            then create a route to the NAT gateway of the designed
            public subnet.
        :type nat_to: None | str
        """
        # Create the subnet
        self.add(SubnetStack(name, self.vpc, cidr_block))

        if is_public:
            # Public subnet
            # Connect to the internet
            self[name].add(
                Route(
                    name + "InternetRoute",
                    self[name].route_table,
                    "0.0.0.0/0",
                    self.gateway,
                    self.gate_attach,
                )
            )
            if use_nat:
                # Add if needed a NAT gateway
                self[name].add(EIP(name + "NatEIP", self.gate_attach))
                self[name].add(
                    NatGateway(
                        name + "NatGateway",
                        self[name][name + "NatEIP"],
                        self[name].subnet,
                    )
                )
        elif nat_to:
            assert nat_to in self, "invalid subnet name: %s" % nat_to
            assert nat_to + "NatGateway" in self[nat_to], (
                "subnet %s has no NAT gateway" % nat_to
            )
            self[name].add(
                Route(
                    name + "NatRoute",
                    self[name].route_table,
                    "0.0.0.0/0",
                    self[nat_to][nat_to + "NatGateway"],
                    self.gate_attach,
                )
            )

    @property
    def vpc(self):
        """Get the VPC CloudFormation resource."""
        return self[self.name]

    @property
    def gateway(self):
        """Get the Gateway CloudFormation resource."""
        return self[self.name + "InternetGateway"]

    @property
    def gate_attach(self):
        """Get the GateAttachment CloudFormation resource."""
        return self[self.name + "GateLink"]


class Fortress(Stack):
    def __init__(
        self,
        name,
        internal_server_policy,
        bastion_ami=None,
        allow_ssh_from=None,
        description=None,
        vpc_cidr_block="10.10.0.0/16",
        private_cidr_block="10.10.0.0/17",
        public_cidr_block="10.10.128.0/18",
        aws_endpoints_cidr_block="10.10.192.0/18",
    ):
        """Create a VPC Fortress.

        This create a vpc with a public and a private subnet. Servers in the
        private subnet are only accessible through a bastion machine declared
        in the public subnet. An additional subnet is created to host AWS
        services endpoints network interfaces.

        :param name: stack name
        :type name: str
        :param internal_server_policy: policy associated with instance role
            of private servers
        :type internal_server_policy: Policy
        :param bastion_ami: AMI used for the bastion server. If None no bastion
            is setup
        :type bastion_ami: AMI | None
        :param allow_ssh_from: ip ranges from which ssh can be done to the
            bastion. if bastion_ami is None, parameter is discarded
        :type allow_ssh_from: str | None
        :param vpc_cidr_block: ip ranges for the associated vpc
        :type vpc_cidr_block: str
        :param private_cidr_block: ip ranges (subset of vpc_cidr_block) used
            for private subnet
        :type private_cidr_block: str
        :param public_cidr_block: ip ranges (subset of vpc_cidr_block) used
            for public subnet
        :type public_cidr_block: str
        :param aws_endpoints_cidr_block: ip ranges (subset of vpc_cidr_block) used
            for aws endpoints
        :type aws_endpoints_cidr_block: str
        """
        super().__init__(name, description)

        # Create VPC along with the three subnets
        self.add(VPCStack(self.name + "VPC", vpc_cidr_block))
        self.vpc.add_subnet(
            self.name + "PublicNet", public_cidr_block, is_public=True, use_nat=True
        )
        self.vpc.add_subnet(
            self.name + "PrivateNet", private_cidr_block, nat_to=self.name + "PublicNet"
        )
        self.vpc.add_subnet(self.name + "AWSEndpointsNet", aws_endpoints_cidr_block)

        self.amazon_groups = {}
        self.github_groups = {}

        if bastion_ami is not None:
            # Allow ssh to bastion only from a range of IP address
            self.add(
                SecurityGroup(
                    self.name + "BastionSG",
                    self.vpc.vpc,
                    description="security group for bastion servers",
                    rules=[Ipv4IngressRule("ssh", cidr) for cidr in allow_ssh_from],
                )
            )

            # Create the bastion
            self.add(Instance(self.name + "Bastion", bastion_ami))
            self.bastion.tags["Name"] = "Bastion (%s)" % self.name
            self.bastion.add(
                EC2NetworkInterface(
                    self.public_subnet.subnet,
                    public_ip=True,
                    groups=[self[self.name + "BastionSG"]],
                )
            )

            # Create security group for internal servers
            self.add(
                SecurityGroup(
                    self.name + "InternalSG",
                    self.vpc.vpc,
                    description=(
                        "Allow ssh inside VPC and allow https "
                        "to VPC endpoints subnet"
                    ),
                    rules=[
                        Ipv4IngressRule("ssh", self.public_subnet.cidr_block),
                        Ipv4EgressRule("https", self.aws_endpoints_subnet.cidr_block),
                    ],
                )
            )
        else:
            # If no bastion is used do not authorize ssh inside the vpc
            self.add(
                SecurityGroup(
                    self.name + "InternalSG",
                    self.vpc.vpc,
                    description=(
                        "Do not allow ssh inside VPC but allow https "
                        "to the VPC endpoints subnet."
                    ),
                    rules=[
                        Ipv4EgressRule("https", self.aws_endpoints_subnet.cidr_block)
                    ],
                )
            )

        # Create security group for endpoints
        self.add(
            SecurityGroup(
                self.name + "InterfaceEndpointsSG",
                self.vpc.vpc,
                description=("Allow https from the private subnet"),
                rules=[Ipv4IngressRule("https", self.private_subnet.cidr_block)],
            )
        )

        ir = InstanceRole(self.name + "PrivServerInstanceRole")
        ir.add_policy(internal_server_policy)
        self.add(ir)

    @property
    def region(self):
        """Return the region in which the stack is allocated.

        :return: a region
        :rtype: str
        """
        return self[self.name + "VPC"].region

    def add_service_access(
        self,
        service_name: str,
        policy_document: PolicyDocument | None = None,
        endpoint_name: str | None = None,
    ) -> None:
        """Add an interface endpoint for a given service.

        :param service_name: name of the service for which to add an interface endpoint
        :param policy_document: optional policy to limit resources that are
            accessibles through the endpoint.
        :param endpont_name: name of the endpoint
        """
        if not endpoint_name:
            endpoint_name = f"{self.name}{service_name}EndPoint"
        if endpoint_name not in self:
            self.add(
                VPCInterfaceEndpoint(
                    name=endpoint_name,
                    service=service_name,
                    vpc=self.vpc,
                    subnet=self.aws_endpoints_subnet.subnet,
                    policy_document=policy_document,
                    security_group=self.aws_endpoints_security_group,
                )
            )

    def add_lambda_access(self, lambda_arns):
        """Add a lambda interface endpoint with permissions to invoke given lambdas.

        :param lambda_arns: arn identifying the lambda to give access to
        :type lambda_arns: list[str]
        """
        endpoint_name = f"{self.name}LambdaEndPoint"
        pd = PolicyDocument()
        pd.append(
            Allow(
                to=["lambda:InvokeFunction"],
                on=chain.from_iterable(
                    ((lambda_arn, lambda_arn + ":*") for lambda_arn in lambda_arns)
                ),
                apply_to=Principal(PrincipalKind.EVERYONE),
            )
        )
        self.add_service_access(
            service_name="lambda", policy_document=pd, endpoint_name=endpoint_name
        )

    def add_secret_access(self, secret_arn):
        """Give read access to a given secret.

        :param secret_name: arn identifying the secret to give access to
        :type secret_name: str
        """
        endpoint_name = f"{self.name}SecretsManagerEndPoint"
        pd = PolicyDocument()
        pd.append(
            Allow(
                to=[
                    "secretsmanager:GetResourcePolicy",
                    "secretsmanager:GetSecretValue",
                    "secretsmanager:DescribeSecret",
                    "secretsmanager:ListSecretVersionIds",
                ],
                on=[secret_arn],
                apply_to=Principal(PrincipalKind.EVERYONE),
            )
        )
        self.add_service_access(
            service_name="secretsmanager",
            policy_document=pd,
            endpoint_name=endpoint_name,
        )

    def add_network_access(
        self,
        protocol: str,
        cidr_block: str = "0.0.0.0/0",
        from_port: int | None = None,
        to_port: int | None = None,
    ) -> None:
        """Authorize some outbound protocols for internal servers.

        :param protocol: protocol name
        :param cidr_block: allowed IP range (default is all)
        :param from_port: optional starting port
        :param to_port: optional ending port
        """
        self[self.name + "InternalSG"].add_rule(
            Ipv4EgressRule(protocol, cidr_block, from_port=from_port, to_port=to_port)
        )

    def add_s3_endpoint_access(self):
        self[self.name + "InternalSG"].add_rule(
            PrefixListEgressRule(
                "https", PREFIX_LISTS[self[self.name + "InternalSG"].region]["s3"]
            )
        )

    def private_server_security_groups(
        self, amazon_access=True, github_access=True, extra_groups=None
    ):
        """Return list of security groups to apply to private servers.

        :param amazon_access: if True add a security group that allow access to
            amazon services. Default is True
        :type amazon_access: bool
        :param github_access: if True add a security group that allow access to
            github services. Default is True
        :type github_access: bool
        :param extra_groups: additional security groups
        :type extra_groups: Optional[list[SecurityGroups]]
        :return: a list of security groups
        :rtype: list[SecurityGroups]
        """
        groups = [self[self.name + "InternalSG"]]
        if amazon_access:
            if not self.amazon_groups:
                self.amazon_groups = amazon_security_groups(
                    self.name + "AmazonServices", self.vpc.vpc
                )
                for sg in self.amazon_groups.values():
                    self.add(sg)

            for group in self.amazon_groups.values():
                groups.append(group)

        if github_access:
            if not self.github_groups:
                self.github_groups = github_security_groups(
                    name=f"{self.name}GitHub", vpc=self.vpc.vpc, protocol="ssh"
                )
                for sg in self.github_groups.values():
                    self.add(sg)

            for group in self.github_groups.values():
                groups.append(group)

        if extra_groups:
            # Register the groups if necesssary
            for group in extra_groups:
                if group.name not in self:
                    self.add(group)
            groups += extra_groups
        return groups

    def add_private_server(
        self,
        server_ami,
        names,
        instance_type="t2.micro",
        disk_size=None,
        amazon_access=True,
        github_access=False,
        persistent_eni=False,
        is_template=False,
        template_name=None,
        extra_groups=None,
    ):
        """Add servers in the private network.

        :param server_ami: AMI to use
        :type server_ami: AMI
        :param names: list of server names (names will be used as stack logical
            names)
        :type names: list[str]
        :param instance_type: instance type (default: t2.micro)
        :type instance_type: str
        :param disk_size: disk size of the instance in Go or None to reuse the
            AMI snapshot size
        :type disk_size: int | None
        :param amazon_access: if True add a security group that allow access to
            amazon services. Default is True
        :type amazon_access: bool
        :param github_access: if True add a security group that allow access to
            github services. Default is False
        :type github_access: bool
        :param persistent_eni: Use a separate network interface (i.e: not
            embedded inside the EC2 instance). This is useful to preserve for
            example IP address and MAC address when a server is redeployed.
        :type persistent_eni: bool
        :param is_template: create a template rather than an instance
        :type is_template: bool
        :param extra_groups: a list of security groups to add
        :type extra_groups: Optional[list[SecurityGroup]]
        """
        groups = self.private_server_security_groups(
            amazon_access=amazon_access,
            github_access=github_access,
            extra_groups=extra_groups,
        )
        nb_sg_groups = len(groups)
        if nb_sg_groups > 16:
            raise AWSFortressError(
                f"Number of security groups is {nb_sg_groups} and exceeds the maximum "
                "number of 16 security groups allowed per network interface."
            )

        for name in names:
            if not is_template:
                self.add(
                    Instance(
                        name,
                        server_ami,
                        instance_type=instance_type,
                        disk_size=disk_size,
                    )
                )
            else:
                self.add(
                    LaunchTemplate(
                        name,
                        server_ami,
                        instance_type=instance_type,
                        disk_size=disk_size,
                        template_name=template_name,
                    )
                )

            if not persistent_eni:
                self[name].add(
                    EC2NetworkInterface(
                        self.private_subnet.subnet, public_ip=False, groups=groups
                    )
                )
            else:
                network_interface = NetworkInterface(
                    name + "ENI", subnet=self.private_subnet.subnet, groups=groups
                )
                self.add(network_interface)
                self[name].add(EC2NetworkInterface(interface=network_interface))

            self[name].set_instance_profile(
                self[self.name + "PrivServerInstanceRole"].instance_profile
            )
            self[name].tags["Name"] = "%s (%s)" % (name, self.name)

    @property
    def vpc(self):
        return self[self.name + "VPC"]

    @property
    def private_subnet(self):
        return self.vpc[self.name + "PrivateNet"]

    @property
    def aws_endpoints_subnet(self):
        return self.vpc[self.name + "AWSEndpointsNet"]

    @property
    def internal_security_group(self):
        return self[self.name + "InternalSG"]

    @property
    def aws_endpoints_security_group(self):
        return self[self.name + "InterfaceEndpointsSG"]

    @property
    def secretsmanager_endpoint(self):
        return self[self.name + "SecretsManagerEndPoint"]

    @property
    def lambda_endpoint(self):
        return self[self.name + "LambdaEndPoint"]

    @property
    def public_subnet(self):
        return self.vpc[self.name + "PublicNet"]

    @property
    def bastion(self):
        return self[self.name + "Bastion"]



"""Provide AWS Config construct tests."""

from e3.aws.troposphere.config.configuration_recorder import ConfigurationRecorder
from e3.aws.troposphere.config.config_rule import (
    S3BucketPublicWriteProhibited,
    S3BucketPublicReadProhibited,
    S3BucketServerSideEncryptionEnabled,
    S3BucketSSLRequestsOnly,
    IAMUserNoPoliciesCheck,
)
from e3.aws.troposphere import Stack

EXPECTED_RULES = {
    "S3BucketPublicWriteProhibited": {
        "Properties": {
            "ConfigRuleName": "s3-bucket-public-write-prohibited",
            "Description": (
                "Checks that your S3 buckets do not allow public write access."
                "If an S3 bucket policy or bucket ACL allows public write access"
                ", the bucket is noncompliant."
            ),
            "InputParameters": {},
            "Scope": {"ComplianceResourceTypes": ["AWS::S3::Bucket"]},
            "Source": {
                "Owner": "AWS",
                "SourceIdentifier": "S3_BUCKET_PUBLIC_WRITE_PROHIBITED",
            },
        },
        "Type": "AWS::Config::ConfigRule",
        "DependsOn": "ConfigRecorder",
    },
    "S3BucketPublicReadProhibited": {
        "Properties": {
            "ConfigRuleName": "s3-bucket-public-read-prohibited",
            "Description": (
                "Checks that your S3 buckets do not allow public read access."
                "If an S3 bucket policy or bucket ACL allows public read access,"
                " the bucket is noncompliant."
            ),
            "InputParameters": {},
            "Scope": {"ComplianceResourceTypes": ["AWS::S3::Bucket"]},
            "Source": {
                "Owner": "AWS",
                "SourceIdentifier": "S3_BUCKET_PUBLIC_READ_PROHIBITED",
            },
        },
        "Type": "AWS::Config::ConfigRule",
        "DependsOn": "ConfigRecorder",
    },
    "S3BucketServerSideEncryptionEnabled": {
        "Properties": {
            "ConfigRuleName": "s3-bucket-server-side-encryption-enabled",
            "Description": (
                "Checks that your Amazon S3 bucket either has S3 default "
                "encryption enabled or that the S3 bucket policy explicitly "
                "denies put-object requests without server side encryption."
            ),
            "InputParameters": {},
            "Scope": {"ComplianceResourceTypes": ["AWS::S3::Bucket"]},
            "Source": {
                "Owner": "AWS",
                "SourceIdentifier": "S3_BUCKET_SERVER_SIDE_ENCRYPTION_ENABLED",
            },
        },
        "Type": "AWS::Config::ConfigRule",
        "DependsOn": "ConfigRecorder",
    },
    "S3BucketSslRequestsOnly": {
        "Properties": {
            "ConfigRuleName": "s3-bucket-ssl-requests-only",
            "Description": (
                "Checks whether S3 buckets have policies that require requests "
                "to use Secure Socket Layer (SSL)."
            ),
            "InputParameters": {},
            "Scope": {"ComplianceResourceTypes": ["AWS::S3::Bucket"]},
            "Source": {
                "Owner": "AWS",
                "SourceIdentifier": "S3_BUCKET_SSL_REQUESTS_ONLY",
            },
        },
        "Type": "AWS::Config::ConfigRule",
        "DependsOn": "ConfigRecorder",
    },
    "IamUserNoPoliciesCheck": {
        "Properties": {
            "ConfigRuleName": "iam-user-no-policies-check",
            "Description": (
                "Checks that none of your IAM users have policies attached. "
                "IAM users must inherit permissions from IAM groups or roles."
            ),
            "InputParameters": {},
            "Scope": {"ComplianceResourceTypes": ["AWS::IAM::User"]},
            "Source": {
                "Owner": "AWS",
                "SourceIdentifier": "IAM_USER_NO_POLICIES_CHECK",
            },
        },
        "Type": "AWS::Config::ConfigRule",
        "DependsOn": "ConfigRecorder",
    },
}

EXPECTED_RECORDER = {
    "AWSServiceRoleForConfig": {
        "Properties": {"AWSServiceName": "config.amazonaws.com"},
        "Type": "AWS::IAM::ServiceLinkedRole",
    },
    "ConfigRecorder": {
        "Properties": {
            "Name": "ConfigRecorder",
            "RecordingGroup": {
                "AllSupported": True,
                "IncludeGlobalResourceTypes": True,
            },
            "RoleARN": {
                "Fn::Join": [
                    ":",
                    [
                        "arn",
                        "aws",
                        "iam:",
                        {"Ref": "AWS::AccountId"},
                        "role/aws-service-role/config.amazonaws.com/"
                        "AWSServiceRoleForConfig",
                    ],
                ]
            },
        },
        "Type": "AWS::Config::ConfigurationRecorder",
        "DependsOn": "AWSServiceRoleForConfig",
    },
    "ConfigTestBucket": {
        "Properties": {
            "BucketName": "config-test-bucket",
            "BucketEncryption": {
                "ServerSideEncryptionConfiguration": [
                    {"ServerSideEncryptionByDefault": {"SSEAlgorithm": "AES256"}}
                ]
            },
            "PublicAccessBlockConfiguration": {
                "BlockPublicAcls": True,
                "BlockPublicPolicy": True,
                "IgnorePublicAcls": True,
                "RestrictPublicBuckets": True,
            },
            "VersioningConfiguration": {"Status": "Enabled"},
        },
        "Type": "AWS::S3::Bucket",
    },
    "ConfigTestBucketPolicy": {
        "Properties": {
            "Bucket": {"Ref": "ConfigTestBucket"},
            "PolicyDocument": {
                "Version": "2012-10-17",
                "Statement": [
                    {
                        "Effect": "Deny",
                        "Principal": {"AWS": "*"},
                        "Action": "s3:*",
                        "Resource": "arn:aws:s3:::config-test-bucket/*",
                        "Condition": {"Bool": {"aws:SecureTransport": "false"}},
                    },
                    {
                        "Effect": "Deny",
                        "Principal": {"AWS": "*"},
                        "Action": "s3:PutObject",
                        "Resource": "arn:aws:s3:::config-test-bucket/*",
                        "Condition": {
                            "StringNotEquals": {
                                "s3:x-amz-server-side-encryption": "AES256"
                            }
                        },
                    },
                    {
                        "Effect": "Deny",
                        "Principal": {"AWS": "*"},
                        "Action": "s3:PutObject",
                        "Resource": "arn:aws:s3:::config-test-bucket/*",
                        "Condition": {
                            "Null": {"s3:x-amz-server-side-encryption": "true"}
                        },
                    },
                    {
                        "Effect": "Allow",
                        "Principal": {"Service": "config.amazonaws.com"},
                        "Action": "s3:GetBucketAcl",
                        "Resource": "arn:aws:s3:::config-test-bucket",
                    },
                    {
                        "Effect": "Allow",
                        "Principal": {"Service": "config.amazonaws.com"},
                        "Action": "s3:PutObject",
                        "Resource": {
                            "Fn::Join": [
                                "",
                                [
                                    "arn:aws:s3:::config-test-bucket",
                                    "/AWSLogs/",
                                    {"Ref": "AWS::AccountId"},
                                    "/Config/*",
                                ],
                            ]
                        },
                        "Condition": {
                            "StringEquals": {
                                "s3:x-amz-acl": "bucket-owner-full-control"
                            }
                        },
                    },
                ],
            },
        },
        "Type": "AWS::S3::BucketPolicy",
    },
    "DeliveryChannel": {
        "Properties": {
            "Name": "DeliveryChannel",
            "S3BucketName": {"Ref": "ConfigTestBucket"},
        },
        "Type": "AWS::Config::DeliveryChannel",
    },
}


def test_config_recorder(stack: Stack) -> None:
    """Test config recorder creation."""
    stack.add(ConfigurationRecorder(bucket_name="config-test-bucket"))
    assert stack.export()["Resources"] == EXPECTED_RECORDER


def test_config_rules(stack: Stack) -> None:
    """Test config rules creation."""
    for config_rule in (
        S3BucketPublicWriteProhibited,
        S3BucketPublicReadProhibited,
        S3BucketServerSideEncryptionEnabled,
        S3BucketSSLRequestsOnly,
        IAMUserNoPoliciesCheck,
    ):
        stack.add(config_rule)

    assert stack.export()["Resources"] == EXPECTED_RULES



#!/usr/bin/env python
# strip .tox/*/lib/python*/site-packages paths from coverage data
# show only paths corresponding to the real source files
# From https://github.com/danilobellini/pytest-doctest-custom/


import os
import sys

from coverage.sqldata import CoverageData
from coverage.files import PathAliases
from tempfile import NamedTemporaryFile


def fix_paths(site_pkg_dir: str, cov_data_file: str) -> None:
    site_pkg_dir = os.path.abspath(site_pkg_dir)

    paths = PathAliases()
    paths.add(site_pkg_dir, "src")

    old_cov_file = NamedTemporaryFile()
    old_cov_file.close()
    os.rename(cov_data_file, old_cov_file.name)

    old_coverage_data = CoverageData(old_cov_file.name)
    old_coverage_data.read()
    new_coverage_data = CoverageData(cov_data_file)
    new_coverage_data.update(old_coverage_data, aliases=paths)
    new_coverage_data.write()


if __name__ == "__main__":
    fix_paths(sys.argv[1], sys.argv[2])



from __future__ import annotations
from enum import Enum
from functools import cached_property
from typing import TYPE_CHECKING
from abc import abstractmethod
from e3.aws import name_to_id
from e3.aws.troposphere import Construct
from troposphere import (
    apigatewayv2,
    route53,
    Ref,
    logs,
    GetAtt,
    awslambda,
    Sub,
    apigateway,
)
from troposphere.apigateway import BasePathMapping
from troposphere.apigatewayv2 import ApiMapping
from e3.aws.troposphere.iam.policy_document import PolicyDocument
from e3.aws.troposphere.iam.policy_statement import PolicyStatement, Trust
from e3.aws.troposphere.iam.role import Role
from troposphere import AWSObject
from troposphere.certificatemanager import Certificate, DomainValidationOption
import json
import re

if TYPE_CHECKING:
    from e3.aws.troposphere import Stack
    from typing import Any, TypedDict


class AuthorizationType(Enum):
    """Allowed authorization types for ApiGateway routes."""

    NONE = "NONE"
    JWT = "JWT"
    IAM = "AWS_IAM"
    CUSTOM = "CUSTOM"


# Declare some constants to make declarations more concise.
NO_AUTH = AuthorizationType.NONE
JWT_AUTH = AuthorizationType.JWT
IAM_AUTH = AuthorizationType.IAM
CUSTOM_AUTH = AuthorizationType.CUSTOM


class Method:
    """API Gateway method definition."""

    def __init__(
        self,
        method: str,
        auth: AuthorizationType | None = None,
        authorizer_name: str | None = None,
    ) -> None:
        """Initialize an API Gateway method definition.

        :param method: the https method
        :param auth: the authorization type associated with the method
        :param authorizer_name: the name of the authorizer to use
            (used only when using JWT_AUTH)
        """
        self.method = method
        self.auth = auth if auth is not None else NO_AUTH
        self.authorizer_name = authorizer_name


class Route(Method):
    """API Gateway route definition."""

    def __init__(
        self,
        method: str,
        route: str,
        auth: AuthorizationType = NO_AUTH,
        authorizer_name: str | None = None,
    ) -> None:
        """Initialize an API Gateway route definition.

        :param method: the https method
        :param route: the route (should start with a "/")
        :param auth: the authorization type associated with the route
        :param authorizer_name: the name of the authorizer to use
            (used only when using JWT_AUTH)
        """
        assert route.startswith("/"), "route path should starts with a /"
        self.route = route
        super().__init__(method=method, auth=auth, authorizer_name=authorizer_name)


class GET(Route):
    """An API Gateway GET route."""

    def __init__(
        self,
        route: str,
        auth: AuthorizationType = NO_AUTH,
        authorizer_name: str | None = None,
    ) -> None:
        """Initialize a GET route.

        :param route: the route (should start with a "/")
        :param auth: the authorization type associated with the route
        :param authorizer_name: the name of the authorizer to use
            (used only when using JWT_AUTH)
        """
        super().__init__(
            method="GET", route=route, auth=auth, authorizer_name=authorizer_name
        )


class POST(Route):
    """An API Gateway POST route."""

    def __init__(
        self,
        route: str,
        auth: AuthorizationType = NO_AUTH,
        authorizer_name: str | None = None,
    ) -> None:
        """Initialize a POST route.

        :param route: the route (should start with a "/")
        :param auth: the authorization type associated with the route
        :param authorizer_name: the name of the authorizer to use
            (used only when using JWT_AUTH)
        """
        super().__init__(
            method="POST", route=route, auth=auth, authorizer_name=authorizer_name
        )


class StageConfiguration(object):
    """HTTP API stage configuration."""

    def __init__(
        self,
        name: str,
        api_mapping_key: str | None = None,
        lambda_arn_permission: str | GetAtt | Ref | None = None,
        variables: dict[str, str] | None = None,
    ) -> None:
        """Create a stage configuration.

        :param name: name of the stage (use $default for the default stage)
        :param api_mapping_key: the API mapping key (only used when creating an HTTP API
            with a domain name and hosted zone id)
        :param lambda_arn_permission: lambda arn for which to add InvokeFunction
            permission for this stage (can be different from the lambda arn executed
            by the HTTP API)
        :param variables: a map that defines the stage variables
        """
        self.name = name
        self.api_mapping_key = api_mapping_key
        self.lambda_arn_permission = lambda_arn_permission
        self.variables = variables


class Api(Construct):
    """API abstact Class for APIGateways V1 and V2."""

    if TYPE_CHECKING:

        class _AliasTargetAttributes(TypedDict):
            DNSName: str
            HostedZoneId: str

    def __init__(
        self,
        name: str,
        description: str,
        lambda_arn: str | GetAtt | Ref,
        burst_limit: int = 10,
        rate_limit: int = 10,
        domain_name: str | None = None,
        hosted_zone_id: str | None = None,
        stages_config: list[StageConfiguration] | None = None,
    ):
        """Initialize API resource.

        :param name: the resource name
        :param description: the resource description
        :param lambda_arn: arn of the lambda executed for all routes
        :param burst_limit: maximum concurrent requests at a given time
            (exceeding that limit will cause API Gateway to return 429)
        :param rate_limit: maximum number of requests per seconds
        :param domain_name: if domain_name is not None then associate the API
            with a given domain name. In that case a certificate is
            automatically created for that domain name. Note that if a domain
            name is specified then the default endpoint (execute-api) is
            disabled.
        :param hosted_zone_id: id of the hosted zone that contains domain_name.
            This parameter is required if domain_name is not None
        :param stages_config: configurations of the different stages
        :param integration_uri: URI of a Lambda function
        """
        self.name = name
        self.description = description
        self.lambda_arn = lambda_arn
        self.burst_limit = burst_limit
        self.rate_limit = rate_limit
        self.domain_name = domain_name
        self.disable_execute_api_endpoint: bool = False
        if self.domain_name is not None:
            self.disable_execute_api_endpoint = True
            assert (
                hosted_zone_id is not None
            ), "hosted zone id required when domain_name is not None"
        self.hosted_zone_id = hosted_zone_id
        self.authorizers: dict[str, Any] = {}
        # By default, make sure to have a $default stage
        self.stages_config = (
            stages_config if stages_config else [StageConfiguration("$default")]
        )

    @cached_property
    def logical_id(self) -> str:
        """Get the API's logical ID."""
        return name_to_id(self.name)

    def stage_logical_id(self, stage_name: str) -> str:
        """Get the name of the Stage resource.

        :param stage_name: the stage name
        """
        return self.logical_id + name_to_id(stage_name) + "Stage"

    @cached_property
    def log_group_id(self) -> str:
        """Get logGroup's id."""
        return self.logical_id + "LogGroup"

    @cached_property
    def ref(self) -> Ref:
        """Return ref to the Gateway API."""
        return Ref(self.logical_id)

    def stage_ref(self, stage_name: str) -> Ref:
        """Return ref to one of the Gateway API stage.

        :param stage_name: the stage name
        """
        return Ref(self.stage_logical_id(stage_name))

    @abstractmethod
    def add_jwt_authorizer(self, name: str, header: str = "Authorization") -> None:
        """Declare a JWT authorizer.

        :param name: authorizer name
        :param header: the request header holding the authorization token
            submitted by the client
        """
        pass

    def cfn_policy_document(self, stack: Stack) -> PolicyDocument:
        """Get policy needed by CloudFormation."""
        return PolicyDocument(
            [
                PolicyStatement(
                    action=["logs:DescribeLogGroups", "logs:CreateLogGroup"],
                    effect="Allow",
                    resource=f"arn:aws:logs:::log-group:{self.name}",
                )
            ]
        )

    @abstractmethod
    def declare_stage(
        self,
        stage_name: str,
        log_arn: str | GetAtt,
        *,
        stage_variables: dict[str, str] | None = None,
    ) -> apigatewayv2.Stage | list[AWSObject]:
        """Declare an API gateway stage.

        :param stage_name: name of the stage
        :param log_arn: arn of the cloudwatch log group in which api calls
            should be logged
        :param stage_variables: variables for the different stages
        :return: the AWSObject corresponding to the Stage
        """
        pass

    def _declare_certificate(
        self, domain_name: str, hosted_zone_id: str
    ) -> Certificate:
        """Declare the API's domain certificate.

        :param domain_name: domain name
        :param hosted_zone_id: hosted zone in which the domain belongs to
        :return: domain name certificate
        """
        return Certificate(
            name_to_id(self.name + domain_name + "Certificate"),
            DomainName=domain_name,
            DomainValidationOptions=[
                DomainValidationOption(
                    DomainName=domain_name, HostedZoneId=hosted_zone_id
                )
            ],
            ValidationMethod="DNS",
        )

    @abstractmethod
    def _declare_domain_name(
        self, domain_name: str, certificate_arn: Ref | str
    ) -> apigatewayv2.DomainName | apigateway.DomainName:
        """Declare the domain name aws resource of the API.

        :param domain_name: domain name
        :param certificate_arn: the ARN of the certificate
        :return: the domain name aws resource
        """
        pass

    @abstractmethod
    def _declare_api_mapping(
        self, domain_name: apigatewayv2.DomainName | apigateway.DomainName
    ) -> list[BasePathMapping | ApiMapping]:
        """Declare the API's mapping.

        :param domain_name: the custom domain name for the API
        return: a list api mapping aws object
        """
        pass

    @abstractmethod
    def _get_alias_target_attributes(self) -> Api._AliasTargetAttributes:
        """Get atributes to pass to GetAtt for alias target."""
        pass

    def declare_domain(self, domain_name: str, hosted_zone_id: str) -> list[AWSObject]:
        """Declare a custom domain for the API stages.

        Note that when a custom domain is created then a certificate is automatically
        created for that domain.

        :param domain_name: domain name
        :param hosted_zone_id: hosted zone in which the domain belongs to
        :return: a list of AWSObject
        """
        result = []

        certificate = self._declare_certificate(
            domain_name=domain_name, hosted_zone_id=hosted_zone_id
        )
        result.append(certificate)

        domain = self._declare_domain_name(
            domain_name=domain_name,
            certificate_arn=certificate.ref(),
        )
        result.append(domain)

        result += self._declare_api_mapping(domain)
        alias_target = self._get_alias_target_attributes()

        result.append(
            route53.RecordSetType(
                name_to_id(self.name + domain_name + "DNS"),
                Name=domain_name,
                Type="A",
                HostedZoneId=hosted_zone_id,
                AliasTarget=route53.AliasTarget(
                    DNSName=GetAtt(
                        domain.title,
                        alias_target["DNSName"],
                    ),
                    HostedZoneId=GetAtt(
                        domain.title,
                        alias_target["HostedZoneId"],
                    ),
                    EvaluateTargetHealth=False,
                ),
            )
        )
        return result


class HttpApi(Api):
    """HTTP API support."""

    def __init__(
        self,
        name: str,
        description: str,
        lambda_arn: str | GetAtt | Ref,
        route_list: list[Route],
        burst_limit: int = 10,
        rate_limit: int = 10,
        domain_name: str | None = None,
        hosted_zone_id: str | None = None,
        stages_config: list[StageConfiguration] | None = None,
        integration_uri: str | Ref | Sub | None = None,
    ):
        """Initialize an HTTP API.

        The schema supported here is a single lambda handling all the routes.
        Nevertherless, we don't use {proxy}+ route and rather declare
        statically the list of supported route. This ensure that our lambda is
        not executed whenever an invalid route is invoked. Thus the cost of invoking
        an invalid route is 0.

        The API can use one stage ($default) or multiple stages with stages_config.

        :param name: the resource name
        :param description: the resource description
        :param lambda_arn: arn of the lambda executed for all routes
        :param route_list: a list of route to declare
        :param burst_limit: maximum concurrent requests at a given time
            (exceeding that limit will cause API Gateway to return 429)
        :param rate_limit: maximum number of requests per seconds
        :param domain_name: if domain_name is not None then associate the API
            with a given domain name. In that case a certificate is
            automatically created for that domain name. Note that if a domain
            name is specified then the default endpoint (execute-api) is
            disabled.
        :param hosted_zone_id: id of the hosted zone that contains domain_name.
            This parameter is required if domain_name is not None
        :param stages_config: configurations of the different stages
        :param integration_uri: URI of a Lambda function
        """
        super().__init__(
            name=name,
            description=description,
            lambda_arn=lambda_arn,
            burst_limit=burst_limit,
            rate_limit=rate_limit,
            domain_name=domain_name,
            hosted_zone_id=hosted_zone_id,
            stages_config=stages_config,
        )
        self.route_list = route_list
        self.integration_uri = (
            integration_uri if integration_uri is not None else lambda_arn
        )

    def add_jwt_authorizer(  # type: ignore
        self, name: str, audience: list[str], issuer: str, header: str = "Authorization"
    ) -> None:
        """Declare a JWT authorizer.

        :param name: authorizer name
        :param audience: a list of accepted audience. For most cases this is the list
            of accepted Cognito client ids.
        :param issuer: the base domain of the entity provider. If Cognito is used this
            is https://cognito-idp.{region}.amazonaws.com/{userPoolId}
        """
        self.authorizers[name] = {
            "ApiId": self.ref,
            "AuthorizerType": "JWT",
            "Name": name,
            "IdentitySource": [f"$request.header.{header}"],
            "JwtConfiguration": apigatewayv2.JWTConfiguration(
                Audience=audience, Issuer=issuer
            ),
        }

    def declare_stage(
        self,
        stage_name: str,
        log_arn: str | GetAtt,
        *,
        stage_variables: dict[str, str] | None = None,
    ) -> apigatewayv2.Stage:
        """Declare an API gateway stage.

        :param stage_name: name of the stage
        :param log_arn: arn of the cloudwatch log group in which api calls
            should be logged
        :param stage_variables: variables for the different stages
        :return: the AWSObject corresponding to the Stage
        """
        log_format = {
            "source_ip": "$context.identity.sourceIp",
            "request_time": "$context.requestTime",
            "method": "$context.httpMethod",
            "route": "$context.routeKey",
            "protocol": "$context.protocol",
            "status": "$context.status",
            "response_length": "$context.responseLength",
            "request_id": "$context.requestId",
            "integration_error_msg": "$context.integrationErrorMessage",
        }

        access_log_settings = apigatewayv2.AccessLogSettings(
            DestinationArn=GetAtt(self.log_group_id, "Arn"),
            Format=json.dumps(log_format),
        )

        route_settings = apigatewayv2.RouteSettings(
            DetailedMetricsEnabled=True,
            ThrottlingBurstLimit=self.burst_limit,
            ThrottlingRateLimit=self.rate_limit,
        )

        parameters: dict[str, Any] = {}

        if stage_variables is not None:
            parameters["StageVariables"] = stage_variables

        return apigatewayv2.Stage(
            self.stage_logical_id(stage_name),
            AccessLogSettings=access_log_settings,
            ApiId=Ref(self.logical_id),
            AutoDeploy=True,
            Description=f"stage {stage_name}",
            DefaultRouteSettings=route_settings,
            StageName=stage_name,
            **parameters,
        )

    def declare_route(self, route: Route, integration: Ref | str) -> list[AWSObject]:
        """Declare a route.

        :param route: the route definition
        :param integration: arn of the integration to use for this route
        :return: a list of AWSObjects to be added to the stack
        """
        result = []
        id_prefix = name_to_id(self.name + route.method + route.route)

        route_params = {
            "ApiId": self.ref,
            "AuthorizationType": route.auth.value,
            "RouteKey": f"{route.method} {route.route}",
            "Target": Sub(
                "integrations/${integration}", dict_values={"integration": integration}
            ),
        }
        if route.authorizer_name:
            route_params["AuthorizerId"] = Ref(name_to_id(route.authorizer_name))

        result.append(apigatewayv2.Route(id_prefix + "Route", **route_params))

        for config in self.stages_config:
            result.append(
                awslambda.Permission(
                    # Retain old behavior for the $default stage
                    name_to_id(
                        "{}-{}LambdaPermission".format(
                            id_prefix, "" if config.name == "$default" else config.name
                        )
                    ),
                    Action="lambda:InvokeFunction",
                    FunctionName=config.lambda_arn_permission
                    if config.lambda_arn_permission is not None
                    else self.lambda_arn,
                    Principal="apigateway.amazonaws.com",
                    SourceArn=Sub(
                        "arn:aws:execute-api:${AWS::Region}:${AWS::AccountId}:"
                        f"${{api}}/{config.name}/${{route_arn}}",
                        dict_values={
                            "api": self.ref,
                            "route_arn": f"{route.method}{route.route}",
                        },
                    ),
                )
            )
        return result

    def _declare_domain_name(
        self, domain_name: str, certificate_arn: Ref | str
    ) -> apigatewayv2.DomainName | apigateway.DomainName:
        """Declare the domain name aws resource of the API.

        :param domain_name: domain name
        :param certificate_arn: the ARN of the certificate
        :return: a domain name aws resource
        """
        return apigatewayv2.DomainName(
            name_to_id(self.name + domain_name + "Domain"),
            DomainName=domain_name,
            DomainNameConfigurations=[
                apigatewayv2.DomainNameConfiguration(CertificateArn=certificate_arn)
            ],
        )

    def _declare_api_mapping(
        self, domain_name: apigatewayv2.DomainName | apigateway.DomainName
    ) -> list[BasePathMapping | ApiMapping]:
        """Declare the API's mapping.

        :param domain_name: the custom domain name for the API
        return: a list api mapping aws object
        """
        result = []
        for config in self.stages_config:
            mapping_params = {
                "DomainName": domain_name.ref(),
                "Stage": self.stage_ref(config.name),
                "ApiId": self.ref,
            }

            if config.api_mapping_key is not None:
                mapping_params["ApiMappingKey"] = config.api_mapping_key
            result.append(
                apigatewayv2.ApiMapping(
                    # Retain old behavior for the $default stage
                    name_to_id(
                        "{}{}-{}ApiMapping".format(
                            self.name,
                            domain_name.DomainName,
                            "" if config.name == "$default" else config.name,
                        )
                    ),
                    **mapping_params,
                )
            )
        return result

    def _get_alias_target_attributes(self) -> Api._AliasTargetAttributes:
        """Get atributes to pass to GetAtt for alias target."""
        return {
            "DNSName": "RegionalDomainName",
            "HostedZoneId": "RegionalHostedZoneId",
        }

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return list of AWSObject associated with the construct."""
        result = []

        # Create a log group for the API
        result.append(logs.LogGroup(self.log_group_id, LogGroupName=self.name))

        # Create the API itself
        api_params = {
            "Description": self.description,
            "ProtocolType": "HTTP",
            "Name": self.name,
            "DisableExecuteApiEndpoint": self.disable_execute_api_endpoint,
        }
        result.append(apigatewayv2.Api(self.logical_id, **api_params))

        # Declare the different stages
        for config in self.stages_config:
            result.append(
                self.declare_stage(
                    stage_name=config.name,
                    log_arn=GetAtt(self.log_group_id, "Arn"),
                    stage_variables=config.variables,
                )
            )

        # Declare one integration
        result.append(
            apigatewayv2.Integration(
                self.logical_id + "Integration",
                ApiId=self.ref,
                IntegrationType="AWS_PROXY",
                IntegrationUri=self.integration_uri,
                PayloadFormatVersion="2.0",
            )
        )

        # Declare the routes
        for route in self.route_list:
            result += self.declare_route(
                route=route, integration=Ref(self.logical_id + "Integration")
            )

        # Declare the domain
        if self.domain_name is not None:
            assert self.hosted_zone_id is not None
            result += self.declare_domain(
                domain_name=self.domain_name, hosted_zone_id=self.hosted_zone_id
            )

        # Declare the authorizers
        for auth_name, auth_params in self.authorizers.items():
            result.append(apigatewayv2.Authorizer(name_to_id(auth_name), **auth_params))

        return result


class RestApi(Api):
    """Rest API support."""

    def __init__(
        self,
        name: str,
        description: str,
        lambda_arn: str | GetAtt | Ref,
        method_list: list[Method],
        burst_limit: int = 10,
        rate_limit: int = 10,
        domain_name: str | None = None,
        hosted_zone_id: str | None = None,
        stages_config: list[StageConfiguration] | None = None,
        integration_uri: str | Ref | Sub | None = None,
        iam_path: str = "/",
        policy: list[PolicyStatement] | None = None,
    ):
        """Initialize a Rest API.

        The schema supported here is a single lambda handling the API.
        To limit lambda invocation costs we can add a policy that restricts
        invocation.

        The API can use either only one stage ($default) or multiple.

        For the moment the RestApi construct has a bug: the base URL
        does not work properly and returns an error message when used
        even if a "/" route is set in the lambda. For example, if we
        set as the domain name of the REST API example.com, trying to
        use or access  the URL example.com returns the following error
        message:  '{"message":"Missing Authentication Token"}' even if
        an authentication method has not been set. However, using other
        routes should work fine e.g. example.com/hello. So for the moment
        using the default(/) in the lambda should be avoided.

        :param name: the resource name
        :param lambda_arn: arn of the lambda executed for all routes
        :param method_list: a list of methods to declare
        :param burst_limit: maximum concurrent requests at a given time
            (exceeding that limit will cause API Gateway to return 429)
        :param rate_limit: maximum number of requests per seconds
        :param domain_name: if domain_name is not None then associate the API
            with a given domain name. In that case a certificate is
            automatically created for that domain name. Note that if a domain
            name is specified then the default endpoint (execute-api) is
            disabled.
        :param hosted_zone_id: id of the hosted zone that contains domain_name.
            This parameter is required if domain_name is not None
        :param stages_config: configurations of the different stages
        :param integration_uri: URI of a Lambda function
        :param iam_path: IAM path for cloudwatch permission and role
            (must be either / or a string starting and ending with /)
        :param policy: the policy document that contains the permissions for
            the RestApi resource.
        """
        super().__init__(
            name=name,
            description=description,
            lambda_arn=lambda_arn,
            burst_limit=burst_limit,
            rate_limit=rate_limit,
            domain_name=domain_name,
            hosted_zone_id=hosted_zone_id,
            stages_config=stages_config,
        )
        self.method_list = method_list
        self.integration_uri = (
            integration_uri
            if integration_uri is not None
            else Sub(
                "arn:aws:apigateway:${AWS::Region}:lambda:path/2015-03-31"
                "/functions/${lambdaArn}/invocations",
                dict_values={"lambdaArn": lambda_arn},
            )
        )
        assert iam_path.startswith("/"), "iam_path must start with '/'"
        assert iam_path.endswith("/"), "iam_path must end with '/'"
        self.iam_path = iam_path
        self.policy = policy

    def add_jwt_authorizer(  # type: ignore
        # we ignore the incompatible signature mypy errors
        self,
        name: str,
        providers_arn: list[str],
        header: str = "Authorization",
    ) -> None:
        """Declare a JWT authorizer.

        :param name: authorizer name
        :param providers_arn: a  list of the Cognito user pool ARNs for the
            COGNITO_USER_POOLS authorizer.Each element is of this format:
            arn:aws:cognito-idp:{region}:{account_id}:userpool/{user_pool_id}.
        :param header: the request header holding the authorization token
            submitted by the client
        """
        self.authorizers[name] = {
            "AuthorizerUri": self.integration_uri,
            "IdentitySource": f"$request.header.{header}",
            "Name": name,
            "ProviderARNs": providers_arn,
            "RestApiId": self.ref,
            "Type": "COGNITO_USER_POOLS",
        }

    def declare_stage(
        self,
        stage_name: str,
        log_arn: str | GetAtt,
        *,
        stage_variables: dict[str, str] | None = None,
    ) -> list[AWSObject]:
        """Declare an API gateway stage.

        :param stage_name: name of the stage
        :param log_arn: arn of the cloudwatch log group in which api calls
            should be logged
        :param stage_variables: variables for the different stages
        :return: Stage and Deployment AWSObjects
        """
        result = []

        # create deployment resource
        deployment_name = self.logical_id + name_to_id(stage_name) + "Deployment"
        result.append(
            apigateway.Deployment(
                deployment_name,
                Description=f"Deployment resource of {stage_name} stage",
                RestApiId=Ref(self.logical_id),
                DependsOn=[
                    name_to_id(self.name + method.method + "Method")
                    for method in self.method_list
                ],
            )
        )

        # create stage resource

        log_format = {
            "source_ip": "$context.identity.sourceIp",
            "request_time": "$context.requestTime",
            "method": "$context.httpMethod",
            "route": "$context.routeKey",
            "protocol": "$context.protocol",
            "status": "$context.status",
            "response_length": "$context.responseLength",
            "request_id": "$context.requestId",
            "integration_error_msg": "$context.integrationErrorMessage",
        }

        access_log_settings = apigateway.AccessLogSetting(
            DestinationArn=GetAtt(self.log_group_id, "Arn"),
            Format=json.dumps(log_format),
        )

        method_settings = apigateway.MethodSetting(
            ResourcePath="/*",
            HttpMethod="*",
            MetricsEnabled=True,
            ThrottlingBurstLimit=self.burst_limit,
            ThrottlingRateLimit=self.rate_limit,
        )

        parameters: dict[str, Any] = {}

        if stage_variables is not None:
            parameters["Variables"] = stage_variables

        result.append(
            apigateway.Stage(
                self.stage_logical_id(stage_name=stage_name),
                AccessLogSetting=access_log_settings,
                RestApiId=Ref(self.logical_id),
                DeploymentId=Ref(deployment_name),
                Description=f"stage {stage_name}",
                MethodSettings=[method_settings],
                # Stage name only allows a-zA-Z0-9_
                StageName=re.sub("[^a-zA-Z0-9_]", "", stage_name),
                **parameters,
            )
        )

        return result

    def declare_method(self, method: Method, resource_id: Ref) -> list[AWSObject]:
        """Declare a method.

        :param method: the method definition
        :return: a list of AWSObjects to be added to the stack
        """
        result = []
        id_prefix = name_to_id(self.name + method.method)

        integration = apigateway.Integration(
            id_prefix + "Integration",
            # set at POST because we are doing lambda integration
            CacheKeyParameters=[],
            CacheNamespace="none",
            IntegrationHttpMethod="POST",
            PassthroughBehavior="NEVER",
            Type="AWS_PROXY",
            Uri=self.integration_uri,
        )

        method_params = {
            "RestApiId": self.ref,
            "AuthorizationType": "COGNITO_USER_POOLS"
            if method.authorizer_name
            else "NONE",
            "HttpMethod": f"{method.method}",
            "Integration": integration,
            "ResourceId": resource_id,
        }
        if method.authorizer_name:
            method_params["AuthorizerId"] = Ref(name_to_id(method.authorizer_name))

        result.append(apigateway.Method(id_prefix + "Method", **method_params))

        result.append(
            awslambda.Permission(
                id_prefix + "LambdaPermission",
                Action="lambda:InvokeFunction",
                FunctionName=self.lambda_arn,
                Principal="apigateway.amazonaws.com",
                SourceArn=Sub(
                    "arn:aws:execute-api:${AWS::Region}:${AWS::AccountId}:" "${api}/*",
                    dict_values={
                        "api": self.ref,
                    },
                ),
            )
        )
        return result

    def _declare_domain_name(
        self, domain_name: str, certificate_arn: Ref | str
    ) -> apigatewayv2.DomainName | apigateway.DomainName:
        """Declare the domain name aws resource of the API.

        :param domain_name: domain name
        :param certificate_arn: the ARN of the certificate
        :return: a domain name aws resource
        """
        return apigateway.DomainName(
            name_to_id(self.name + domain_name + "Domain"),
            DomainName=domain_name,
            CertificateArn=certificate_arn,
        )

    def _declare_api_mapping(
        self, domain_name: apigatewayv2.DomainName | apigateway.DomainName
    ) -> list[BasePathMapping | ApiMapping]:
        """Declare the API's mapping.

        :param domain_name: the custom domain name for the API
        return: a list api mapping aws object
        """
        result = []
        for config in self.stages_config:
            mapping_params = {
                "DomainName": domain_name.ref(),
                "Stage": self.stage_ref(config.name),
                "RestApiId": self.ref,
            }

            if config.api_mapping_key is not None:
                mapping_params["BasePath"] = config.api_mapping_key
            result.append(
                apigateway.BasePathMapping(
                    # Retain old behavior for the $default stage
                    name_to_id(
                        "{}{}-{}BasePathMapping".format(
                            self.name,
                            domain_name.DomainName,
                            "" if config.name == "$default" else config.name,
                        )
                    ),
                    **mapping_params,
                )
            )
        return result

    def _get_alias_target_attributes(self) -> Api._AliasTargetAttributes:
        """Get atributes to pass to GetAtt for alias target."""
        return {
            "DNSName": "DistributionDomainName",
            "HostedZoneId": "DistributionHostedZoneId",
        }

    def declare_access_cloudwatch_resources(self) -> list[AWSObject]:
        """Create role and account resources to enable CloudWatch Logs.

        Specify the IAM role that Amazon API Gateway should use to write API
        logs to Amazon CloudWatch.These resources must be created at least
        once per account per region for the class RestApi to be able to be
        deployed.

        :return: a role and an account cloudformation objects
        """
        result: list[AWSObject] = []

        access_cloudwatch_role = Role(
            name="EnableCloudwatchLoggingForApigatewayRole",
            description="role that enables CloudWatch logging for apigateway v1"
            " apigateway",
            trust=Trust(services=["apigateway"]),
            managed_policy_arns=[
                # we use the AmazonAPIGatewayPushToCloudWatchLogs managed policy
                # that has all the required permissions
                Sub(
                    "arn:${AWS::Partition}:iam::aws:policy/service-role/AmazonAPIGatewayPushToCloudWatchLogs"
                )
            ],
            path=self.iam_path,
        )

        result.append(access_cloudwatch_role)
        result.append(
            apigateway.Account(
                self.logical_id + "Account",
                CloudWatchRoleArn=access_cloudwatch_role.arn,
            )
        )

        return result

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return list of AWSObject associated with the construct."""
        result = []

        # Create a log group for the API
        result.append(logs.LogGroup(self.log_group_id, LogGroupName=self.name))

        # Create the API itself
        api_params = {
            "Description": self.description,
            "Name": self.name,
            "DisableExecuteApiEndpoint": self.disable_execute_api_endpoint,
        }
        if self.policy:
            api_params["Policy"] = PolicyDocument(statements=self.policy).as_dict

        result.append(apigateway.RestApi(self.logical_id, **api_params))

        # Create an API resource
        resource_name = self.logical_id + "Resource"
        result.append(
            apigateway.Resource(
                resource_name,
                ParentId=GetAtt(self.logical_id, "RootResourceId"),
                RestApiId=self.ref,
                PathPart="{proxy+}",
            )
        )

        # Declare the different stages
        for config in self.stages_config:
            result.extend(
                self.declare_stage(
                    stage_name=config.name,
                    log_arn=GetAtt(self.log_group_id, "Arn"),
                    stage_variables=config.variables,
                )
            )

        # Declare the methods
        for method in self.method_list:
            result += self.declare_method(method=method, resource_id=Ref(resource_name))

        # Declare the domain
        if self.domain_name is not None:
            assert self.hosted_zone_id is not None
            result += self.declare_domain(
                domain_name=self.domain_name, hosted_zone_id=self.hosted_zone_id
            )

        # Declare the authorizers
        for auth_name, auth_params in self.authorizers.items():
            result.append(apigateway.Authorizer(name_to_id(auth_name), **auth_params))

        return result



from enum import Enum

from e3.aws.cfn import AWSType, Resource
from e3.aws.cfn.iam import PolicyDocument


class AccessControl(Enum):
    """Canned ACLs for buckets."""

    AUTHENTICATED_READ = "AuthenticatedRead"
    AWS_EXEC_READ = "AwsExecRead"
    BUCKET_OWNER_READ = "BucketOwnerRead"
    BUCKET_OWNER_FULL_CONTROL = "BucketOwnerFullControl"
    LOG_DELIVERY_WRITE = "LogDeliveryWrite"
    PRIVATE = "Private"
    PUBLIC_READ = "PublicRead"
    PUBLIC_WRITE = "PublicReadWrite"


class BucketPolicy(Resource):
    """S3 Bucket Policy."""

    def __init__(self, name, bucket, policy_document):
        """Initialize a bucket policy.

        :param name: logical name of the resource in the stack
        :type name: str
        :param bucket: bucket on which to apply the policy
        :type bucket: Bucket
        :param policy_document: Policy document to apply
        :type policy_document: PolicyDocument
        """
        super(BucketPolicy, self).__init__(name, kind=AWSType.S3_BUCKET_POLICY)
        assert isinstance(policy_document, PolicyDocument)
        assert isinstance(bucket, str) or isinstance(bucket, Bucket)
        self.bucket = bucket
        self.policy_document = policy_document

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        result = {"PolicyDocument": self.policy_document.properties}
        if isinstance(self.bucket, Bucket):
            result["Bucket"] = self.bucket.ref
        else:
            result["Bucket"] = self.bucket
        return result


class Bucket(Resource):
    """S3 Bucket."""

    ATTRIBUTES = ("Arn", "DomainName")

    def __init__(self, name, access_control=None, bucket_name=None, versioning=False):
        """Initialize a S3 bucket.

        :param name: logical name of the resource in the stack
        :type name: str
        :param acccess_control: A canned access control list (ACL) that grants
            predefined permissions to the bucket. if None default is PRIVATE
        :type access_control: None | AccessControl
        :param bucket_name: the bucket name in AWS. If set cloud formation will
            not be able to update settings of the buckets automatically.
        :type bucket_name: str | None
        """
        super(Bucket, self).__init__(name, kind=AWSType.S3_BUCKET)
        if access_control is None:
            self.access_control = AccessControl.PRIVATE
        else:
            assert isinstance(access_control, AccessControl)
            self.access_control = access_control
        self.bucket_name = bucket_name
        self.versioning = versioning

    @property
    def arn(self):
        return self.getatt("Arn")

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        result = {
            "AccessControl": self.access_control.value,
            "BucketEncryption": {
                "ServerSideEncryptionConfiguration": [
                    {"ServerSideEncryptionByDefault": {"SSEAlgorithm": "aws:kms"}}
                ]
            },
        }
        if self.bucket_name is not None:
            result["BucketName"] = self.bucket_name
        if self.versioning:
            result["VersioningConfiguration"] = {"Status": "Enabled"}
        else:
            result["VersioningConfiguration"] = {"Status": "Suspended"}
        return result



from __future__ import annotations
from typing import TYPE_CHECKING

from e3.aws.troposphere import Construct, Stack, name_to_id
from e3.aws.troposphere.iam.managed_policy import ManagedPolicy
from e3.aws.troposphere.iam.policy_statement import Allow

from troposphere import secretsmanager, Ref

if TYPE_CHECKING:
    from troposphere import AWSObject
    from e3.aws.troposphere.awslambda import Function


class Secret(Construct):
    """Provide Secret resource and associated utility methods."""

    def __init__(self, name: str, description: str, iam_path: str = "/"):
        """Initialize a Secret instance.

        :param name: name of the secret
        :param description: secret description
        :param iam_path: IAM path for the rotation policy
        """
        self.name = name
        self.description = description
        self.iam_path = iam_path
        self.secret = secretsmanager.Secret(
            name_to_id(name), Description=description, Name=name
        )

    @property
    def rotation_lambda_policy(self) -> ManagedPolicy:
        """Return policy granting permissions to rotate the secret."""
        return ManagedPolicy(
            name=name_to_id(f"{self.name}RotationPolicy"),
            description="Managed policy granting permissions"
            f"to rotate the {self.name} secret",
            statements=[
                Allow(
                    action=[
                        "secretsmanager:DescribeSecret",
                        "secretsmanager:GetSecretValue",
                        "secretsmanager:PutSecretValue",
                        "secretsmanager:UpdateSecretVersionStage",
                    ],
                    resource=self.secret_arn,
                )
            ],
            path=self.iam_path,
        )

    @property
    def secret_arn(self) -> str:
        """Return secret arn."""
        return Ref(self.secret)

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return troposphere objects associated with the construct."""
        return [self.secret]


class RotationSchedule(Construct):
    """Provide resources to schedule rotation for a secret."""

    def __init__(
        self, secret: Secret, rotation_function: Function, schedule_expression: str
    ):
        """Initialize a ScheduleRotationSecret instance.

        :param secret: secret to rotate
        :param rotation_function: lambda function that rotates the secret
        :param schedule_expression: expression that schedule the rotation.
            It has to follow the scheduled event rules format
            https://docs.aws.amazon.com/AmazonCloudWatch/latest/events/ScheduledEvents.html
        """
        self.schedule_expression = schedule_expression
        self.lambda_permission = rotation_function.invoke_permission(
            name_suffix="permission",
            service="secretsmanager",
            source_arn=secret.secret_arn,
        )
        self.schedule = secretsmanager.RotationSchedule(
            name_to_id(f"{secret.name}RotationSchedule"),
            RotationLambdaARN=rotation_function.arn,
            RotationRules=secretsmanager.RotationRules(
                ScheduleExpression=schedule_expression
            ),
            SecretId=secret.secret_arn,
            DependsOn=self.lambda_permission.title,
        )

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return troposphere objects associated with the construct."""
        return [self.lambda_permission, self.schedule]



from __future__ import annotations
import base64
import docker
import os
import pytest
import json

from troposphere.awslambda import (
    ProvisionedConcurrencyConfiguration,
    AliasRoutingConfiguration,
    VersionWeight,
)

from e3.aws import AWSEnv
from e3.aws.troposphere import Stack
from e3.aws.troposphere.awslambda import (
    PyFunction,
    Py38Function,
    DockerFunction,
    Alias,
    Version,
    AutoVersion,
    BlueGreenAliases,
    BlueGreenAliasConfiguration,
)
from e3.aws.troposphere.awslambda.flask_apigateway_wrapper import FlaskLambdaHandler

SOURCE_DIR = os.path.join(os.path.dirname(os.path.abspath(__file__)), "source_dir")


EXPECTED_PY38FUNCTION_TEMPLATE = {
    "Mypylambda": {
        "Properties": {
            "Code": {
                "S3Bucket": "cfn_bucket",
                "S3Key": "templates/mypylambda_lambda.zip",
            },
            "Description": "this is a test",
            "FunctionName": "mypylambda",
            "Handler": "app.main",
            "Role": "somearn",
            "Runtime": "python3.8",
            "Timeout": 3,
        },
        "Type": "AWS::Lambda::Function",
    }
}

EXPECTED_PYFUNCTION_DEFAULT_TEMPLATE = {
    "Mypylambda": {
        "Properties": {
            "Code": {
                "S3Bucket": "cfn_bucket",
                "S3Key": "templates/mypylambda_lambda.zip",
            },
            "Description": "this is a test",
            "FunctionName": "mypylambda",
            "Handler": "app.main",
            "Role": "somearn",
            "Runtime": "python3.9",
            "Timeout": 3,
            "EphemeralStorage": {"Size": 4096},
            "Environment": {
                "Variables": {"env_key_1": "env_value_1", "env_key_2": "env_value2"}
            },
        },
        "Type": "AWS::Lambda::Function",
    },
    "MypylambdaLogGroup": {
        "DeletionPolicy": "Retain",
        "Properties": {
            "LogGroupName": "/aws/lambda/mypylambda",
            "RetentionInDays": 731,
        },
        "Type": "AWS::Logs::LogGroup",
    },
}

EXPECTED_PYFUNCTION_TEMPLATE = {
    "Mypylambda": {
        "Properties": {
            "Code": {
                "S3Bucket": "cfn_bucket",
                "S3Key": "templates/mypylambda_lambda.zip",
            },
            "Description": "this is a test",
            "FunctionName": "mypylambda",
            "Handler": "app.main",
            "Role": "somearn",
            "Runtime": "python3.9",
            "Timeout": 3,
            "MemorySize": 128,
            "EphemeralStorage": {"Size": 1024},
            "ReservedConcurrentExecutions": 1,
            "Environment": {
                "Variables": {"env_key_1": "env_value_1", "env_key_2": "env_value2"}
            },
        },
        "Type": "AWS::Lambda::Function",
    },
    "MypylambdaLogGroup": {
        "DeletionPolicy": "Retain",
        "Properties": {
            "LogGroupName": "/aws/lambda/mypylambda",
            "RetentionInDays": 7,
        },
        "Type": "AWS::Logs::LogGroup",
    },
}


EXPECTED_PYFUNCTION_POLICY_DOCUMENT = {
    "Statement": [
        {
            "Action": [
                "lambda:CreateFunction",
                "lambda:GetFunction",
                "lambda:DeleteFunction",
                "lambda:UpdateFunctionCode",
                "lambda:UpdateFunctionConfiguration",
                "lambda:GetFunctionConfiguration",
            ],
            "Effect": "Allow",
            "Resource": "arn:aws:lambda:::function:mypylambda*",
        },
        {"Action": ["iam:PassRole"], "Effect": "Allow", "Resource": "somearn"},
    ],
    "Version": "2012-10-17",
}

EXPECTED_DOCKER_FUNCTION = {
    "Dockerfunction": {
        "Properties": {
            "PackageType": "Image",
            "Code": {"ImageUri": "<dry_run_image_uri>"},
            "Timeout": 3,
            "Description": "this is a test",
            "Role": "somearn",
            "FunctionName": "dockerfunction",
        },
        "Type": "AWS::Lambda::Function",
    },
    "DockerfunctionLogGroup": {
        "DeletionPolicy": "Retain",
        "Properties": {
            "LogGroupName": "/aws/lambda/dockerfunction",
            "RetentionInDays": 731,
        },
        "Type": "AWS::Logs::LogGroup",
    },
}

EXPECTED_VERSION_DEFAULT_TEMPLATE = {
    "Prod": {
        "Properties": {
            "Description": "this is the prod version",
            "FunctionName": {"Fn::GetAtt": ["Mypylambda", "Arn"]},
        },
        "Type": "AWS::Lambda::Version",
    }
}

EXPECTED_VERSION_TEMPLATE = {
    "Prod": {
        "Properties": {
            "Description": "this is the prod version",
            "FunctionName": {"Fn::GetAtt": ["Mypylambda", "Arn"]},
            "ProvisionedConcurrencyConfig": {"ProvisionedConcurrentExecutions": 1},
            "CodeSha256": "somesha",
        },
        "Type": "AWS::Lambda::Version",
    }
}

EXPECTED_ALIAS_DEFAULT_TEMPLATE = {
    "Myalias": {
        "Properties": {
            "Name": "myalias",
            "Description": "this is a test",
            "FunctionName": {"Fn::GetAtt": ["Mypylambda", "Arn"]},
            "FunctionVersion": "1",
        },
        "Type": "AWS::Lambda::Alias",
    }
}

EXPECTED_ALIAS_TEMPLATE = {
    "Myalias": {
        "Properties": {
            "Name": "myalias",
            "Description": "this is a test",
            "FunctionName": {"Fn::GetAtt": ["Mypylambda", "Arn"]},
            "FunctionVersion": {"Fn::GetAtt": ["Newversion", "Version"]},
            "ProvisionedConcurrencyConfig": {"ProvisionedConcurrentExecutions": 1},
            "RoutingConfig": {
                "AdditionalVersionWeights": [
                    {
                        "FunctionVersion": {"Fn::GetAtt": ["Oldversion", "Version"]},
                        "FunctionWeight": 0.5,
                    }
                ]
            },
        },
        "Type": "AWS::Lambda::Alias",
    }
}

EXPECTED_AUTOVERSION_DEFAULT_TEMPLATE = {
    "MypylambdaVersion1": {
        "Properties": {
            "Description": "version 1 of mypylambda lambda",
            "FunctionName": {"Fn::GetAtt": ["Mypylambda", "Arn"]},
        },
        "Type": "AWS::Lambda::Version",
    },
    "MypylambdaVersion2": {
        "Properties": {
            "Description": "version 2 of mypylambda lambda",
            "FunctionName": {"Fn::GetAtt": ["Mypylambda", "Arn"]},
        },
        "Type": "AWS::Lambda::Version",
    },
}

EXPECTED_AUTOVERSION_SINGLE_TEMPLATE = {
    "MypylambdaVersion1": {
        "Properties": {
            "Description": "version 1 of mypylambda lambda",
            "FunctionName": {"Fn::GetAtt": ["Mypylambda", "Arn"]},
        },
        "Type": "AWS::Lambda::Version",
    }
}

EXPECTED_AUTOVERSION_TEMPLATE = {
    "MypylambdaVersion2": {
        "Properties": {
            "Description": "version 2 of mypylambda lambda",
            "FunctionName": {"Fn::GetAtt": ["Mypylambda", "Arn"]},
        },
        "Type": "AWS::Lambda::Version",
    },
    "MypylambdaVersion3": {
        "Properties": {
            "Description": "version 3 of mypylambda lambda",
            "FunctionName": {"Fn::GetAtt": ["Mypylambda", "Arn"]},
            "ProvisionedConcurrencyConfig": {"ProvisionedConcurrentExecutions": 1},
            "CodeSha256": "somesha",
        },
        "Type": "AWS::Lambda::Version",
    },
}

EXPECTED_BLUEGREENALIASES_DEFAULT_TEMPLATE = {
    "MypylambdaBlueAlias": {
        "Properties": {
            "Name": "MypylambdaBlueAlias",
            "Description": "blue alias for mypylambda lambda",
            "FunctionName": {"Fn::GetAtt": ["Mypylambda", "Arn"]},
            "FunctionVersion": {"Ref": "MypylambdaVersion1"},
        },
        "Type": "AWS::Lambda::Alias",
    },
    "MypylambdaGreenAlias": {
        "Properties": {
            "Name": "MypylambdaGreenAlias",
            "Description": "green alias for mypylambda lambda",
            "FunctionName": {"Fn::GetAtt": ["Mypylambda", "Arn"]},
            "FunctionVersion": {"Ref": "MypylambdaVersion2"},
        },
        "Type": "AWS::Lambda::Alias",
    },
}

EXPECTED_BLUEGREENALIASES_TEMPLATE = {
    "MypylambdaProdAlias": {
        "Properties": {
            "Name": "MypylambdaProdAlias",
            "Description": "prod alias for mypylambda lambda",
            "FunctionName": {"Fn::GetAtt": ["Mypylambda", "Arn"]},
            "FunctionVersion": {"Ref": "MypylambdaVersion1"},
            "ProvisionedConcurrencyConfig": {"ProvisionedConcurrentExecutions": 1},
            "RoutingConfig": {
                "AdditionalVersionWeights": [
                    {
                        "FunctionVersion": {
                            "Fn::GetAtt": ["MypylambdaVersion1", "Version"]
                        },
                        "FunctionWeight": 1,
                    }
                ]
            },
        },
        "Type": "AWS::Lambda::Alias",
    },
    "MypylambdaBetaAlias": {
        "Properties": {
            "Name": "MypylambdaBetaAlias",
            "Description": "beta alias for mypylambda lambda",
            "FunctionName": {"Fn::GetAtt": ["Mypylambda", "Arn"]},
            "FunctionVersion": {"Ref": "MypylambdaVersion2"},
            "ProvisionedConcurrencyConfig": {"ProvisionedConcurrentExecutions": 1},
            "RoutingConfig": {
                "AdditionalVersionWeights": [
                    {
                        "FunctionVersion": {
                            "Fn::GetAtt": ["MypylambdaVersion2", "Version"]
                        },
                        "FunctionWeight": 1,
                    }
                ]
            },
        },
        "Type": "AWS::Lambda::Alias",
    },
}


@pytest.fixture
def simple_lambda_function() -> PyFunction:
    """Return a simple lambda function for testing."""
    return PyFunction(
        name="mypylambda",
        description="this is a test",
        role="somearn",
        runtime="python3.9",
        code_dir="my_code_dir",
        handler="app.main",
    )


def test_py38function(stack: Stack) -> None:
    """Test Py38Function creation."""
    stack.s3_bucket = "cfn_bucket"
    stack.s3_key = "templates/"
    stack.add(
        Py38Function(
            name="mypylambda",
            description="this is a test",
            role="somearn",
            code_dir="my_code_dir",
            handler="app.main",
        )
    )
    assert stack.export()["Resources"] == EXPECTED_PY38FUNCTION_TEMPLATE


def test_pyfunction_default(stack: Stack) -> None:
    """Test PyFunction creation with default settings."""
    stack.s3_bucket = "cfn_bucket"
    stack.s3_key = "templates/"
    stack.add(
        PyFunction(
            name="mypylambda",
            description="this is a test",
            role="somearn",
            runtime="python3.9",
            code_dir="my_code_dir",
            handler="app.main",
            ephemeral_storage_size=4096,
            environment={"env_key_1": "env_value_1", "env_key_2": "env_value2"},
        )
    )
    print(stack.export()["Resources"])
    assert stack.export()["Resources"] == EXPECTED_PYFUNCTION_DEFAULT_TEMPLATE


def test_pyfunction(stack: Stack) -> None:
    """Test PyFunction creation."""
    stack.s3_bucket = "cfn_bucket"
    stack.s3_key = "templates/"
    stack.add(
        PyFunction(
            name="mypylambda",
            description="this is a test",
            role="somearn",
            runtime="python3.9",
            code_dir="my_code_dir",
            handler="app.main",
            memory_size=128,
            ephemeral_storage_size=1024,
            logs_retention_in_days=7,
            reserved_concurrent_executions=1,
            environment={"env_key_1": "env_value_1", "env_key_2": "env_value2"},
        )
    )
    print(stack.export()["Resources"])
    assert stack.export()["Resources"] == EXPECTED_PYFUNCTION_TEMPLATE


def test_pyfunction_policy_document(stack: Stack) -> None:
    """Test cfn_policy_document of PyFunction."""
    stack.s3_bucket = "cfn_bucket"
    stack.s3_key = "templates/"
    stack.add(
        PyFunction(
            name="mypylambda",
            description="this is a test",
            role="somearn",
            runtime="python3.9",
            code_dir="my_code_dir",
            handler="app.main",
        )
    )
    print(stack.cfn_policy_document().as_dict)
    assert stack.cfn_policy_document().as_dict == EXPECTED_PYFUNCTION_POLICY_DOCUMENT


def test_docker_function(stack: Stack) -> None:
    """Test adding docker function to stack."""
    aws_env = AWSEnv(regions=["us-east-1"], stub=True)
    stubber_ecr = aws_env.stub("ecr")

    stubber_ecr.add_response(
        "get_authorization_token",
        {
            "authorizationData": [
                {
                    "authorizationToken": base64.b64encode(
                        b"test_user:test_pwd"
                    ).decode(),
                    "proxyEndpoint": "test_endpoint",
                }
            ]
        },
        {},
    )

    stack.deploy_session = aws_env
    docker_function = DockerFunction(
        name="dockerfunction",
        description="this is a test",
        role="somearn",
        source_dir=SOURCE_DIR,
        repository_name="e3_aws_test_repository",
        image_tag="test_tag",
    )
    client = docker.from_env()
    try:
        stack.add(docker_function)
    except docker.errors.APIError:
        # Push is expected to fail
        pass
    finally:
        # Always try to remove local test image
        client.images.remove(f"e3_aws_test_repository:{docker_function.image_tag}")

    # Add resources without trying to push the image to ECR
    stack.dry_run = True
    stack.add(docker_function)

    assert stack.export()["Resources"] == EXPECTED_DOCKER_FUNCTION


def test_version_default(stack: Stack, simple_lambda_function: PyFunction) -> None:
    """Test Version creation with default settings."""
    stack.add(
        Version(
            name="prod",
            description="this is the prod version",
            lambda_arn=simple_lambda_function.arn,
        )
    )
    print(stack.export()["Resources"])
    assert stack.export()["Resources"] == EXPECTED_VERSION_DEFAULT_TEMPLATE


def test_version(stack: Stack, simple_lambda_function: PyFunction) -> None:
    """Test Version creation."""
    stack.add(
        Version(
            name="prod",
            description="this is the prod version",
            lambda_arn=simple_lambda_function.arn,
            provisioned_concurrency_config=ProvisionedConcurrencyConfiguration(
                ProvisionedConcurrentExecutions=1
            ),
            code_sha256="somesha",
        )
    )
    print(stack.export()["Resources"])
    assert stack.export()["Resources"] == EXPECTED_VERSION_TEMPLATE


def test_alias_default(stack: Stack, simple_lambda_function: PyFunction) -> None:
    """Test Alias creation with default settings."""
    stack.add(
        Alias(
            name="myalias",
            description="this is a test",
            lambda_arn=simple_lambda_function.arn,
            lambda_version="1",
        )
    )
    print(stack.export()["Resources"])
    assert stack.export()["Resources"] == EXPECTED_ALIAS_DEFAULT_TEMPLATE


def test_alias(stack: Stack, simple_lambda_function: PyFunction) -> None:
    """Test Alias creation."""
    new_version = Version(
        name="newversion",
        description="this is the new version",
        lambda_arn=simple_lambda_function.arn,
    )

    old_version = Version(
        name="oldversion",
        description="this is the old version",
        lambda_arn=simple_lambda_function.arn,
    )

    stack.add(
        Alias(
            name="myalias",
            description="this is a test",
            lambda_arn=simple_lambda_function.arn,
            lambda_version=new_version.version,
            provisioned_concurrency_config=ProvisionedConcurrencyConfiguration(
                ProvisionedConcurrentExecutions=1
            ),
            routing_config=AliasRoutingConfiguration(
                AdditionalVersionWeights=[
                    VersionWeight(
                        FunctionVersion=old_version.version, FunctionWeight=0.5
                    )
                ]
            ),
        )
    )
    print(stack.export()["Resources"])
    assert stack.export()["Resources"] == EXPECTED_ALIAS_TEMPLATE


def test_autoversion_default(stack: Stack, simple_lambda_function: PyFunction) -> None:
    """Test AutoVersion creation with default settings."""
    auto_version = AutoVersion(
        2,
        lambda_function=simple_lambda_function,
    )
    stack.add(auto_version)
    print(stack.export()["Resources"])
    assert stack.export()["Resources"] == EXPECTED_AUTOVERSION_DEFAULT_TEMPLATE
    assert auto_version.get_version(1).name == "mypylambdaVersion1"
    assert auto_version.get_version(2).name == "mypylambdaVersion2"
    assert auto_version.previous.name == "mypylambdaVersion1"
    assert auto_version.latest.name == "mypylambdaVersion2"


def test_autoversion_single(stack: Stack, simple_lambda_function: PyFunction) -> None:
    """Test AutoVersion creation with a single version."""
    auto_version = AutoVersion(
        1,
        lambda_function=simple_lambda_function,
    )
    stack.add(auto_version)
    print(stack.export()["Resources"])
    assert stack.export()["Resources"] == EXPECTED_AUTOVERSION_SINGLE_TEMPLATE
    assert auto_version.previous.name == "mypylambdaVersion1"
    assert auto_version.latest.name == "mypylambdaVersion1"


def test_autoversion(stack: Stack, simple_lambda_function: PyFunction) -> None:
    """Test AutoVersion creation."""
    auto_version = AutoVersion(
        3,
        min_version=2,
        lambda_name=simple_lambda_function.name,
        lambda_arn=simple_lambda_function.arn,
        provisioned_concurrency_config=ProvisionedConcurrencyConfiguration(
            ProvisionedConcurrentExecutions=1
        ),
        code_sha256="somesha",
    )
    stack.add(auto_version)
    print(stack.export()["Resources"])
    assert stack.export()["Resources"] == EXPECTED_AUTOVERSION_TEMPLATE
    assert auto_version.previous.name == "mypylambdaVersion2"
    assert auto_version.latest.name == "mypylambdaVersion3"


def test_bluegreenaliases_default(
    stack: Stack, simple_lambda_function: PyFunction
) -> None:
    """Test BlueGreenAliases creation with default settings."""
    auto_version = AutoVersion(
        2,
        lambda_function=simple_lambda_function,
    )
    aliases = BlueGreenAliases(
        blue_config=BlueGreenAliasConfiguration(version=auto_version.previous.ref),
        green_config=BlueGreenAliasConfiguration(version=auto_version.latest.ref),
        lambda_function=simple_lambda_function,
    )
    stack.add(aliases)
    print(stack.export()["Resources"])
    assert stack.export()["Resources"] == EXPECTED_BLUEGREENALIASES_DEFAULT_TEMPLATE
    assert aliases.blue.name == "MypylambdaBlueAlias"
    assert aliases.green.name == "MypylambdaGreenAlias"


def test_bluegreenaliases(stack: Stack, simple_lambda_function: PyFunction) -> None:
    """Test BlueGreenAliases creation for prod/beta deployment."""
    auto_version = AutoVersion(
        2,
        lambda_function=simple_lambda_function,
    )
    aliases = BlueGreenAliases(
        blue_config=BlueGreenAliasConfiguration(
            version=auto_version.previous.ref,
            name="prod",
            provisioned_concurrency_config=ProvisionedConcurrencyConfiguration(
                ProvisionedConcurrentExecutions=1
            ),
            routing_config=AliasRoutingConfiguration(
                AdditionalVersionWeights=[
                    VersionWeight(
                        FunctionVersion=auto_version.previous.version, FunctionWeight=1
                    )
                ]
            ),
        ),
        green_config=BlueGreenAliasConfiguration(
            version=auto_version.latest.ref,
            name="beta",
            provisioned_concurrency_config=ProvisionedConcurrencyConfiguration(
                ProvisionedConcurrentExecutions=1
            ),
            routing_config=AliasRoutingConfiguration(
                AdditionalVersionWeights=[
                    VersionWeight(
                        FunctionVersion=auto_version.latest.version, FunctionWeight=1
                    )
                ]
            ),
        ),
        lambda_name=simple_lambda_function.name,
        lambda_arn=simple_lambda_function.arn,
    )
    stack.add(aliases)
    print(stack.export()["Resources"])
    assert stack.export()["Resources"] == EXPECTED_BLUEGREENALIASES_TEMPLATE
    assert aliases.blue.name == "MypylambdaProdAlias"
    assert aliases.green.name == "MypylambdaBetaAlias"


def test_create_flask_wsgi_environ_with_rest_api_event():
    # get REST API lambda event
    with open(
        os.path.join(SOURCE_DIR, "event.json"),  # an event from a REST API
    ) as fd:
        rest_api_event = json.load(fd)

    handler = FlaskLambdaHandler("app")
    flask_environment = handler.create_flask_wsgi_environ(rest_api_event, {})

    # remove values that are not
    # JSON serializable
    flask_environment.pop("wsgi.input")
    flask_environment.pop("wsgi.errors")

    # serialize to a JSON dict
    flask_environment = json.loads(json.dumps(flask_environment))

    with open(
        os.path.join(SOURCE_DIR, "rest_api_wsgi_flask_environment.json"),
    ) as fd:
        expected_flask_environment = json.load(fd)

    assert flask_environment == expected_flask_environment



from e3.aws.cfn import AWSType, Resource
import re


class Repository(Resource):
    """CodeCommit Repository."""

    ATTRIBUTES = ("Arn", "CloneUrlHttp", "CloneUrlSsh", "Name")

    def __init__(self, name, description):
        """Initialize a Repository.

        :param name: name of the repository
        :type name: str
        :param description: description of the repository content
        :type description: str
        """
        resource_name = re.sub(r"[^a-zA-Z0-9]+", "", name)
        super(Repository, self).__init__(
            resource_name, kind=AWSType.CODE_COMMIT_REPOSITORY
        )
        self.name = resource_name
        self.repository_name = name
        self.description = description

    @property
    def properties(self):
        """Serialize the object as a simple dict.

        Can be used to transform to CloudFormation Yaml format.

        :rtype: dict
        """
        return {
            "RepositoryName": self.repository_name,
            "RepositoryDescription": self.description,
        }



"""Provide constructs to define aws config rules."""

from __future__ import annotations
from typing import TYPE_CHECKING


from troposphere import AWSObject, AccountId, iam, config, Join

from e3.aws import name_to_id
from e3.aws.troposphere import Construct
from e3.aws.troposphere.s3.bucket import Bucket
from e3.aws.troposphere.iam.policy_statement import PolicyStatement

if TYPE_CHECKING:
    from e3.aws.troposphere import Stack


class ConfigurationRecorder(Construct):
    """Define a ConfigurationRecorder and associated ressources."""

    def __init__(self, bucket_name: str):
        """Initializae a ConfigurationRecorder.

        :param name: The name of the Amazon S3 bucket to create and to which AWS
            Config will deliver configuration snapshots and configuration history
            files.
        """
        self.bucket_name = bucket_name

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Build and return objects associated with the configuration recorder.

        Return a configuration recorder and a delivery channel with its s3 bucket
        """
        aws_objects = []

        config_role = iam.ServiceLinkedRole.from_dict(
            "AWSServiceRoleForConfig", {"AWSServiceName": "config.amazonaws.com"}
        )
        aws_objects.append(config_role)

        # Add the config recorder
        recording_group = config.RecordingGroup(
            AllSupported=True, IncludeGlobalResourceTypes=True
        )

        aws_objects.append(
            config.ConfigurationRecorder(
                name_to_id("ConfigRecorder"),
                Name="ConfigRecorder",
                RecordingGroup=recording_group,
                RoleARN=Join(
                    ":",
                    [
                        "arn",
                        "aws",
                        "iam:",
                        AccountId,
                        (
                            "role/aws-service-role/"
                            "config.amazonaws.com/AWSServiceRoleForConfig"
                        ),
                    ],
                ),
                DependsOn=config_role.title,
            )
        )

        # Create an S3 bucket for the delivery
        bucket = Bucket(name=self.bucket_name)
        bucket.policy_statements += [
            PolicyStatement(
                action="s3:GetBucketAcl",
                effect="Allow",
                principal={"Service": "config.amazonaws.com"},
                resource=bucket.arn,
            ),
            PolicyStatement(
                action="s3:PutObject",
                effect="Allow",
                condition={
                    "StringEquals": {"s3:x-amz-acl": "bucket-owner-full-control"}
                },
                principal={"Service": "config.amazonaws.com"},
                resource=Join("", [bucket.arn, "/AWSLogs/", AccountId, "/Config/*"]),
            ),
        ]

        aws_objects.extend(bucket.resources(stack=stack))

        # Create the delivery channel to the S3 bucket
        aws_objects.append(
            config.DeliveryChannel(
                name_to_id("DeliveryChannel"),
                Name="DeliveryChannel",
                S3BucketName=bucket.ref,
            )
        )

        return aws_objects



from __future__ import annotations
from typing import TYPE_CHECKING
from e3.aws import name_to_id
from e3.aws.troposphere import Construct
from e3.aws.troposphere.iam.policy_document import PolicyDocument
from e3.aws.troposphere.iam.policy_statement import Allow

from troposphere import sns, GetAtt, Ref

if TYPE_CHECKING:
    from typing import Any
    from troposphere import AWSObject
    from e3.aws.troposphere import Stack
    from e3.aws.troposphere.awslambda import Function
    from e3.aws.troposphere.iam.policy_statement import ConditionType


class Topic(Construct):
    """A SNS Topic."""

    def __init__(self, name: str, kms_master_key_id: str | None = None):
        """Initialize a SNS Topic.

        :param name: topic name
        :param kms_master_key_id: the ID of an AWS managed customer master
            key (CMK) for Amazon SNS or a custom CMK
        """
        self.name = name
        self.subscriptions: list[sns.Subscription] = []
        self.optional_resources: list[AWSObject] = []
        self.kms_master_key_id = kms_master_key_id

    def add_lambda_subscription(
        self, function: Function, delivery_policy: dict | None = None
    ) -> None:
        """Add a lambda subscription endpoint to topic.

        :param function: lambda function that will be added as endpoint
        :param delivery_policy: The delivery policy to assign to the subscription
        """
        sub_params = {
            "Endpoint": function.arn,
            "Protocol": "lambda",
            "TopicArn": self.arn,
            "DeliveryPolicy": delivery_policy,
        }

        self.optional_resources.extend(
            [
                sns.SubscriptionResource(
                    name_to_id(f"{function.name}Sub"), **sub_params
                ),
                function.invoke_permission(
                    name_suffix=self.name, service="sns", source_arn=self.arn
                ),
            ]
        )

    def allow_publish_policy(
        self, service: str, name_suffix: str, condition: ConditionType | None = None
    ) -> sns.TopicPolicy:
        """Return a policy allowing a service to publish to the topic.

        :param service: service allowed to publish
        :param name_suffix: a suffix used in the object name
        :param condition: condition to be able to publish
        """
        return sns.TopicPolicy(
            name_to_id(f"{self.name}Policy{name_suffix}"),
            Topics=[self.ref],
            PolicyDocument=PolicyDocument(
                statements=[
                    Allow(
                        action="sns:Publish",
                        resource=self.ref,
                        principal={"Service": f"{service}.amazonaws.com"},
                        condition=condition,
                    )
                ]
            ).as_dict,
        )

    @property
    def arn(self) -> GetAtt:
        """Arn of the SNS Topic."""
        return self.ref

    @property
    def ref(self) -> Ref:
        """Ref of the SNS Topic."""
        return Ref(name_to_id(self.name))

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Compute AWS resources for the construct."""
        params: dict[str, Any] = {}

        if self.kms_master_key_id is not None:
            params["KmsMasterKeyId"] = self.kms_master_key_id

        return [
            sns.Topic(
                name_to_id(self.name),
                TopicName=self.name,
                Subscription=self.subscriptions,
                **params,
            ),
            *self.optional_resources,
        ]



"""Provide ecr construct tests."""

from e3.aws.troposphere import Stack
from e3.aws.troposphere.ecr.repository import Repository

EXPECTED_REPOSITORY = {
    "TestRepository": {
        "Properties": {
            "ImageScanningConfiguration": {"ScanOnPush": True},
            "ImageTagMutability": "IMMUTABLE",
            "RepositoryName": "test-repository",
            "Tags": [{"Key": "Name", "Value": "test-repository"}],
        },
        "Type": "AWS::ECR::Repository",
    }
}


def test_ecr_repository(stack: Stack) -> None:
    """Test ECR repository scheduled rule creation."""
    stack.add(Repository(name="test-repository"))
    assert stack.export()["Resources"] == EXPECTED_REPOSITORY



"""Provide IAM Managed policies."""
from __future__ import annotations
from typing import TYPE_CHECKING
from troposphere import iam, AWSObject, Ref

from e3.aws import name_to_id
from e3.aws.troposphere import Construct
from e3.aws.troposphere.iam.policy_document import PolicyDocument
from e3.aws.troposphere.iam.policy_statement import PolicyStatement, Allow

if TYPE_CHECKING:  # all: no cover
    from e3.aws.troposphere import Stack
    from typing import Any


class ManagedPolicy(Construct):
    """AWS ManagedPolicy."""

    def __init__(
        self,
        name: str,
        statements: list[PolicyStatement],
        description: str = "",
        path: str = "/",
        **kwargs: Any,
    ) -> None:
        """Initialize an IAM Managed policy.

        :param name: name of the managed policy
        :param description: managed_policy description
        :param statements: policy statement part of the policy
        :param path: resource path (either / or a string starting and ending with /)
        :param kwargs: troposphere additional attributes (Condition, DependsOn ...)
        """
        self.name = name
        self.description = description
        self.statements = statements
        self.path = path
        self.kwargs = kwargs

    @property
    def arn(self) -> Ref:
        """Return managed policy arn."""
        return Ref(name_to_id(self.name))

    def resources(self, stack: Stack) -> list[AWSObject]:
        """Return troposphere objects defining the managed policy."""
        params = {
            "Description": self.description,
            "ManagedPolicyName": self.name,
            "PolicyDocument": PolicyDocument(statements=self.statements).as_dict,
            "Path": self.path,
            **self.kwargs,
        }
        return [iam.ManagedPolicy(name_to_id(self.name), **params)]

    def cfn_policy_document(self, stack: Stack) -> PolicyDocument:
        return PolicyDocument(
            statements=[
                Allow(
                    action=[
                        "iam:GetPolicy",
                        "iam:CreatePolicy",
                        "iam:ListPolicyVersions",
                        "iam:DeletePolicy",
                        "iam:CreatePolicyVersion",
                        "iam:DeletePolicyVersion",
                    ],
                    resource=f"arn:aws:iam::%(account)s:policy/{self.name}",
                )
            ]
        )



from __future__ import annotations
import json
import os
import pytest
from e3.aws.troposphere import Stack
from e3.aws.troposphere.awslambda import (
    PyFunction,
    BlueGreenAliases,
    BlueGreenAliasConfiguration,
    AutoVersion,
)
from e3.aws.troposphere.iam.policy_statement import Allow, PolicyStatement
from e3.aws.troposphere.apigateway import (
    JWT_AUTH,
    HttpApi,
    RestApi,
    GET,
    POST,
    Method,
    StageConfiguration,
)

TEST_DIR = os.path.dirname(os.path.abspath(__file__))

# Generic template without any stage
COMMON_TEMPLATE = {
    "Mypylambda": {
        "Properties": {
            "Code": {
                "S3Bucket": "cfn_bucket",
                "S3Key": "templates/mypylambda_lambda.zip",
            },
            "Timeout": 3,
            "Description": "this is a test",
            "Role": "somearn",
            "FunctionName": "mypylambda",
            "Runtime": "python3.8",
            "Handler": "app.main",
        },
        "Type": "AWS::Lambda::Function",
    },
    "TestapiLogGroup": {
        "Properties": {"LogGroupName": "testapi"},
        "Type": "AWS::Logs::LogGroup",
    },
    "Testapi": {
        "Properties": {
            "Description": "this is a test",
            "ProtocolType": "HTTP",
            "Name": "testapi",
            "DisableExecuteApiEndpoint": False,
        },
        "Type": "AWS::ApiGatewayV2::Api",
    },
    "TestapiIntegration": {
        "Properties": {
            "ApiId": {"Ref": "Testapi"},
            "IntegrationType": "AWS_PROXY",
            "IntegrationUri": {"Ref": "Mypylambda"},
            "PayloadFormatVersion": "2.0",
        },
        "Type": "AWS::ApiGatewayV2::Integration",
    },
    "TestapiGETapi1Route": {
        "Properties": {
            "ApiId": {"Ref": "Testapi"},
            "AuthorizationType": "NONE",
            "RouteKey": "GET /api1",
            "Target": {
                "Fn::Sub": [
                    "integrations/${integration}",
                    {"integration": {"Ref": "TestapiIntegration"}},
                ]
            },
        },
        "Type": "AWS::ApiGatewayV2::Route",
    },
    "TestapiPOSTapi2Route": {
        "Properties": {
            "ApiId": {"Ref": "Testapi"},
            "AuthorizationType": "NONE",
            "RouteKey": "POST /api2",
            "Target": {
                "Fn::Sub": [
                    "integrations/${integration}",
                    {"integration": {"Ref": "TestapiIntegration"}},
                ]
            },
        },
        "Type": "AWS::ApiGatewayV2::Route",
    },
}

# Generic stage template with common values
COMMON_STAGE_TEMPLATE = {
    "AccessLogSettings": {
        "DestinationArn": {"Fn::GetAtt": ["TestapiLogGroup", "Arn"]},
        "Format": '{"source_ip": "$context.identity.sourceIp", '
        '"request_time": "$context.requestTime", '
        '"method": "$context.httpMethod", "route": "$context.routeKey", '
        '"protocol": "$context.protocol", "status": "$context.status", '
        '"response_length": "$context.responseLength", '
        '"request_id": "$context.requestId", '
        '"integration_error_msg": "$context.integrationErrorMessage"}',
    },
    "ApiId": {"Ref": "Testapi"},
    "AutoDeploy": True,
    "DefaultRouteSettings": {
        "DetailedMetricsEnabled": True,
        "ThrottlingBurstLimit": 10,
        "ThrottlingRateLimit": 10,
    },
}

# API with a $default stage
EXPECTED_TEMPLATE = {
    **COMMON_TEMPLATE,
    "TestapiGETapi1LambdaPermission": {
        "Properties": {
            "Action": "lambda:InvokeFunction",
            "FunctionName": {"Ref": "Mypylambda"},
            "Principal": "apigateway.amazonaws.com",
            "SourceArn": {
                "Fn::Sub": [
                    "arn:aws:execute-api:${AWS::Region}:"
                    "${AWS::AccountId}:${api}/$default/${route_arn}",
                    {"api": {"Ref": "Testapi"}, "route_arn": "GET/api1"},
                ]
            },
        },
        "Type": "AWS::Lambda::Permission",
    },
    "TestapiPOSTapi2LambdaPermission": {
        "Properties": {
            "Action": "lambda:InvokeFunction",
            "FunctionName": {"Ref": "Mypylambda"},
            "Principal": "apigateway.amazonaws.com",
            "SourceArn": {
                "Fn::Sub": [
                    "arn:aws:execute-api:${AWS::Region}:"
                    "${AWS::AccountId}:${api}/$default/${route_arn}",
                    {"api": {"Ref": "Testapi"}, "route_arn": "POST/api2"},
                ]
            },
        },
        "Type": "AWS::Lambda::Permission",
    },
    "TestapiDefaultStage": {
        "Properties": {
            **COMMON_STAGE_TEMPLATE,
            "Description": "stage $default",
            "StageName": "$default",
        },
        "Type": "AWS::ApiGatewayV2::Stage",
    },
}

# API with $default/beta stages
EXPECTED_TEMPLATE_STAGE = {
    **EXPECTED_TEMPLATE,
    "TestapiGETapi1BetaLambdaPermission": {
        "Properties": {
            "Action": "lambda:InvokeFunction",
            "FunctionName": {"Ref": "Mypylambda"},
            "Principal": "apigateway.amazonaws.com",
            "SourceArn": {
                "Fn::Sub": [
                    "arn:aws:execute-api:${AWS::Region}:${AWS::AccountId}:"
                    "${api}/beta/${route_arn}",
                    {"api": {"Ref": "Testapi"}, "route_arn": "GET/api1"},
                ]
            },
        },
        "Type": "AWS::Lambda::Permission",
    },
    "TestapiPOSTapi2BetaLambdaPermission": {
        "Properties": {
            "Action": "lambda:InvokeFunction",
            "FunctionName": {"Ref": "Mypylambda"},
            "Principal": "apigateway.amazonaws.com",
            "SourceArn": {
                "Fn::Sub": [
                    "arn:aws:execute-api:${AWS::Region}:${AWS::AccountId}:"
                    "${api}/beta/${route_arn}",
                    {"api": {"Ref": "Testapi"}, "route_arn": "POST/api2"},
                ]
            },
        },
        "Type": "AWS::Lambda::Permission",
    },
    "TestapiBetaStage": {
        "Properties": {
            **COMMON_STAGE_TEMPLATE,
            "Description": "stage beta",
            "StageName": "beta",
            "StageVariables": {"somevar": "somevalue"},
        },
        "Type": "AWS::ApiGatewayV2::Stage",
    },
}

# API with $default/beta stages and lambdaAlias variable
EXPECTED_TEMPLATE_LAMBDA_ALIAS = {
    **COMMON_TEMPLATE,
    "TestapiGETapi1LambdaPermission": {
        "Properties": {
            "Action": "lambda:InvokeFunction",
            "FunctionName": {"Ref": "MypylambdaBlueAlias"},
            "Principal": "apigateway.amazonaws.com",
            "SourceArn": {
                "Fn::Sub": [
                    "arn:aws:execute-api:${AWS::Region}:${AWS::AccountId}:"
                    "${api}/$default/${route_arn}",
                    {"api": {"Ref": "Testapi"}, "route_arn": "GET/api1"},
                ]
            },
        },
        "Type": "AWS::Lambda::Permission",
    },
    "TestapiPOSTapi2LambdaPermission": {
        "Properties": {
            "Action": "lambda:InvokeFunction",
            "FunctionName": {"Ref": "MypylambdaBlueAlias"},
            "Principal": "apigateway.amazonaws.com",
            "SourceArn": {
                "Fn::Sub": [
                    "arn:aws:execute-api:${AWS::Region}:${AWS::AccountId}:"
                    "${api}/$default/${route_arn}",
                    {"api": {"Ref": "Testapi"}, "route_arn": "POST/api2"},
                ]
            },
        },
        "Type": "AWS::Lambda::Permission",
    },
    "TestapiGETapi1BetaLambdaPermission": {
        "Properties": {
            "Action": "lambda:InvokeFunction",
            "FunctionName": {"Ref": "MypylambdaGreenAlias"},
            "Principal": "apigateway.amazonaws.com",
            "SourceArn": {
                "Fn::Sub": [
                    "arn:aws:execute-api:${AWS::Region}:${AWS::AccountId}:"
                    "${api}/beta/${route_arn}",
                    {"api": {"Ref": "Testapi"}, "route_arn": "GET/api1"},
                ]
            },
        },
        "Type": "AWS::Lambda::Permission",
    },
    "TestapiPOSTapi2BetaLambdaPermission": {
        "Properties": {
            "Action": "lambda:InvokeFunction",
            "FunctionName": {"Ref": "MypylambdaGreenAlias"},
            "Principal": "apigateway.amazonaws.com",
            "SourceArn": {
                "Fn::Sub": [
                    "arn:aws:execute-api:${AWS::Region}:${AWS::AccountId}:"
                    "${api}/beta/${route_arn}",
                    {"api": {"Ref": "Testapi"}, "route_arn": "POST/api2"},
                ]
            },
        },
        "Type": "AWS::Lambda::Permission",
    },
    "TestapiDefaultStage": {
        "Properties": {
            **COMMON_STAGE_TEMPLATE,
            "Description": "stage $default",
            "StageName": "$default",
            "StageVariables": {"lambdaAlias": "MypylambdaBlueAlias"},
        },
        "Type": "AWS::ApiGatewayV2::Stage",
    },
    "TestapiBetaStage": {
        "Properties": {
            **COMMON_STAGE_TEMPLATE,
            "Description": "stage beta",
            "StageName": "beta",
            "StageVariables": {"lambdaAlias": "MypylambdaGreenAlias"},
        },
        "Type": "AWS::ApiGatewayV2::Stage",
    },
    "TestapiIntegration": {
        "Properties": {
            **EXPECTED_TEMPLATE["TestapiIntegration"]["Properties"],
            "IntegrationUri": "arn:aws:lambda:eu-west-1:123456789012:function:"
            "mypylambda:${stageVariables.lambdaAlias}",
        },
        "Type": "AWS::ApiGatewayV2::Integration",
    },
}


@pytest.fixture
def lambda_fun() -> PyFunction:
    """Return a simple lambda function for testing."""
    return PyFunction(
        name="mypylambda",
        description="this is a test",
        role="somearn",
        code_dir="my_code_dir",
        handler="app.main",
        runtime="python3.8",
        logs_retention_in_days=None,
    )


def test_http_api(stack: Stack, lambda_fun: PyFunction) -> None:
    """Test basic HTTP API."""
    stack.s3_bucket = "cfn_bucket"
    stack.s3_key = "templates/"

    stack.add(lambda_fun)
    stack.add(
        HttpApi(
            name="testapi",
            description="this is a test",
            lambda_arn=lambda_fun.ref,
            route_list=[GET(route="/api1"), POST(route="/api2")],
        )
    )

    assert stack.export()["Resources"] == EXPECTED_TEMPLATE


def test_http_api_stage(stack: Stack, lambda_fun: PyFunction) -> None:
    """Test HTTP API with stages."""
    stack.s3_bucket = "cfn_bucket"
    stack.s3_key = "templates/"

    http_api = HttpApi(
        name="testapi",
        description="this is a test",
        lambda_arn=lambda_fun.ref,
        route_list=[GET(route="/api1"), POST(route="/api2")],
        stages_config=[
            StageConfiguration("$default"),
            StageConfiguration("beta", variables={"somevar": "somevalue"}),
        ],
    )

    stack.add(lambda_fun)
    stack.add(http_api)

    assert stack.export()["Resources"] == EXPECTED_TEMPLATE_STAGE


def test_http_api_lambda_alias(stack: Stack, lambda_fun: PyFunction) -> None:
    """Test HTTP API with lambda alias."""
    stack.s3_bucket = "cfn_bucket"
    stack.s3_key = "templates/"

    lambda_versions = AutoVersion(2, lambda_function=lambda_fun)

    lambda_aliases = BlueGreenAliases(
        blue_config=BlueGreenAliasConfiguration(
            version=lambda_versions.previous.version
        ),
        green_config=BlueGreenAliasConfiguration(
            version=lambda_versions.latest.version
        ),
        lambda_function=lambda_fun,
    )

    http_api = HttpApi(
        name="testapi",
        description="this is a test",
        lambda_arn=lambda_fun.ref,
        route_list=[GET(route="/api1"), POST(route="/api2")],
        stages_config=[
            StageConfiguration(
                "$default",
                lambda_arn_permission=lambda_aliases.blue.ref,
                variables={"lambdaAlias": lambda_aliases.blue.name},
            ),
            StageConfiguration(
                "beta",
                lambda_arn_permission=lambda_aliases.green.ref,
                variables={"lambdaAlias": lambda_aliases.green.name},
            ),
        ],
        integration_uri="arn:aws:lambda:eu-west-1:123456789012:function:"
        f"{lambda_fun.name}:${{stageVariables.lambdaAlias}}",
    )

    stack.add(lambda_fun)
    stack.add(http_api)

    assert stack.export()["Resources"] == EXPECTED_TEMPLATE_LAMBDA_ALIAS


def test_http_api_custom_domain(stack: Stack, lambda_fun: PyFunction) -> None:
    """Test basic HTTP API with custom domain."""
    stack.s3_bucket = "cfn_bucket"
    stack.s3_key = "templates/"

    stack.add(lambda_fun)
    http_api = HttpApi(
        name="testapi",
        description="this is a test",
        lambda_arn=lambda_fun.ref,
        domain_name="api.example.com",
        hosted_zone_id="ABCDEFG",
        route_list=[
            GET(route="/api1"),
            POST(route="/api2"),
            GET("/api3", auth=JWT_AUTH, authorizer_name="testauthorizer"),
        ],
    )
    http_api.add_jwt_authorizer(
        name="testauthorizer",
        audience=["testaudience"],
        issuer="https://cognito-idp.eu-west-1.amazonaws.com/eu-west-1_test",
    )
    stack.add(http_api)

    with open(os.path.join(TEST_DIR, "apigateway_test_custom_domain.json")) as fd:
        expected = json.load(fd)

    print(stack.export()["Resources"])
    assert stack.export()["Resources"] == expected


def test_http_api_multi_domains(stack: Stack, lambda_fun: PyFunction) -> None:
    """Test basic HTTP API with two domains."""
    stack.s3_bucket = "cfn_bucket"
    stack.s3_key = "templates/"

    stack.add(lambda_fun)
    http_api = HttpApi(
        name="testapi",
        description="this is a test",
        lambda_arn=lambda_fun.ref,
        domain_name="api.example.com",
        hosted_zone_id="ABCDEFG",
        route_list=[
            GET(route="/api1"),
            POST(route="/api2"),
        ],
    )
    stack.add(http_api)
    for el in http_api.declare_domain(
        domain_name="api2.example.com", hosted_zone_id="BCDEFGH"
    ):
        stack.add(el)

    with open(os.path.join(TEST_DIR, "apigateway_test_multi_domains.json")) as fd:
        expected = json.load(fd)

    print(stack.export()["Resources"])
    assert stack.export()["Resources"] == expected


def test_http_api_custom_domain_stages(stack: Stack, lambda_fun: PyFunction) -> None:
    """Test basic HTTP API with custom domain and stage."""
    stack.s3_bucket = "cfn_bucket"
    stack.s3_key = "templates/"

    stack.add(lambda_fun)
    http_api = HttpApi(
        name="testapi",
        description="this is a test",
        lambda_arn=lambda_fun.ref,
        domain_name="api.example.com",
        hosted_zone_id="ABCDEFG",
        route_list=[
            GET(route="/api1"),
            POST(route="/api2"),
            GET("/api3", auth=JWT_AUTH, authorizer_name="testauthorizer"),
        ],
        stages_config=[
            StageConfiguration("$default"),
            StageConfiguration("beta", api_mapping_key="beta"),
        ],
    )
    http_api.add_jwt_authorizer(
        name="testauthorizer",
        audience=["testaudience"],
        issuer="https://cognito-idp.eu-west-1.amazonaws.com/eu-west-1_test",
    )
    stack.add(http_api)

    with open(
        os.path.join(TEST_DIR, "apigateway_test_custom_domain_stages.json")
    ) as fd:
        expected = json.load(fd)

    print(stack.export()["Resources"])
    assert stack.export()["Resources"] == expected


def test_rest_api_custom_domain_stages(stack: Stack, lambda_fun: PyFunction) -> None:
    """Test REST api custom domain and stage."""
    stack.s3_bucket = "cfn_bucket"
    stack.s3_key = "templates/"

    stack.add(lambda_fun)
    rest_api = RestApi(
        name="testapi",
        description="this is a test",
        lambda_arn=lambda_fun.ref,
        domain_name="api.example.com",
        hosted_zone_id="ABCDEFG",
        method_list=[
            Method("ANY", authorizer_name="testauthorizer"),
        ],
        stages_config=[
            StageConfiguration("$default"),
            StageConfiguration(
                "beta", api_mapping_key="beta", variables={"somevar": "somevalue"}
            ),
        ],
        policy=[
            Allow(
                principal="*",
                action=[
                    "execute-api:Invoke",
                ],
                resource="execute-api:/*/*/*",
            ),
            # allow API invocation only from a specific IP
            PolicyStatement(
                effect="Deny",
                principal="*",
                action="execute-api:Invoke",
                resource="execute-api:/*/*/*",
                condition={"NotIpAddress": {"aws:SourceIp": ["1.2.3.4"]}},
            ),
        ],
    )
    rest_api.add_jwt_authorizer(
        name="testauthorizer",
        providers_arn=[
            "arn:aws:cognito-idp:eu-west-1:123456789012:userpool/eu-west-1_abc123"
        ],
    )
    stack.add(rest_api)
    with open(
        os.path.join(TEST_DIR, "apigatewayv1_test_custom_domain_stages.json"),
    ) as fd:
        expected = json.load(fd)

    print(stack.export()["Resources"])
    assert stack.export()["Resources"] == expected



"""Provide SecretsManager constructs tests."""
from __future__ import annotations
import json
import os

from e3.aws.troposphere import Stack
from e3.aws.troposphere.secretsmanager import Secret, RotationSchedule
from e3.aws.troposphere.awslambda import PyFunction

TEST_DIR = os.path.dirname(os.path.abspath(__file__))


def test_secret_rotation_schedule(stack: Stack) -> None:
    """Test RotatingSecret."""
    stack.s3_bucket = "cfn_bucket"
    stack.s3_key = "templates/"

    secret = Secret(
        name="TestSecret",
        description="TestSecret description",
    )
    rotation_policy = secret.rotation_lambda_policy
    rotation_function = PyFunction(
        name="myrotationlambda",
        description="this is a test",
        role="somearn",
        runtime="python3.9",
        code_dir="my_code_dir",
        handler="app.main",
    )
    rotation_schedule = RotationSchedule(
        secret=secret,
        rotation_function=rotation_function,
        schedule_expression="rate(4 days)",
    )
    for el in (secret, rotation_policy, rotation_function, rotation_schedule):
        stack.add(el)
    with open(os.path.join(TEST_DIR, "secret_rotation_schedule.json")) as fd:
        expected_template = json.load(fd)

    assert stack.export()["Resources"] == expected_template


def test_secret_iam_path(stack: Stack) -> None:
    """Test Secret with iam_path."""
    secret = Secret(
        name="TestSecret",
        description="TestSecret description",
        iam_path="/iam_test_path/",
    )
    rotation_policy = secret.rotation_lambda_policy
    for el in (secret, rotation_policy):
        stack.add(el)

    with open(os.path.join(TEST_DIR, "secret_iam_path.json")) as fd:
        expected_template = json.load(fd)

    assert stack.export()["Resources"] == expected_template



"""Provide PolicyDocument class."""

from __future__ import annotations
from dataclasses import dataclass
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from e3.aws.troposphere.iam.policy_statement import PolicyStatement


@dataclass
class PolicyDocument:
    """Define a policy document.

    :param statements: policy statements to add to the policy
    :param version: version of language syntax rules that are to be used to
        process the policy
    """

    statements: list[PolicyStatement]
    version: str = "2012-10-17"

    def __add__(self, other: PolicyDocument) -> PolicyDocument:
        """Return a new policy document combining statements from self and other.

        :param other: Other PolicyDocument from which to add statements
        """
        return PolicyDocument(statements=self.statements + other.statements)

    def __iadd__(self, other: PolicyDocument) -> PolicyDocument:
        self.statements += other.statements
        return self

    @property
    def as_dict(self) -> dict:
        """Return dictionary representation of the PolicyDocument."""
        policy_document = {
            "Version": self.version,
            "Statement": [statement.as_dict for statement in self.statements],
        }

        return policy_document



"""Provide InternetGateway construct tests."""

from __future__ import annotations

import pytest
from troposphere import ec2, Ref

from e3.aws import name_to_id
from e3.aws.troposphere import Stack
from e3.aws.troposphere.ec2 import InternetGateway


EXPECTED_TEMPLATE = {
    "VpcTest": {
        "Properties": {
            "CidrBlock": "10.0.0.0/16",
            "EnableDnsHostnames": True,
            "EnableDnsSupport": True,
        },
        "Type": "AWS::EC2::VPC",
    },
    "EuWest1aSubnet": {
        "Properties": {
            "CidrBlock": "10.0.0.0/20",
            "AvailabilityZone": "eu-west-1b",
            "VpcId": {"Ref": "VpcTest"},
            "MapPublicIpOnLaunch": True,
        },
        "Type": "AWS::EC2::Subnet",
    },
    "EuWest1bSubnet": {
        "Properties": {
            "CidrBlock": "10.0.0.0/20",
            "AvailabilityZone": "eu-west-1b",
            "VpcId": {"Ref": "VpcTest"},
            "MapPublicIpOnLaunch": True,
        },
        "Type": "AWS::EC2::Subnet",
    },
    "TestIgw": {"Type": "AWS::EC2::InternetGateway"},
    "TestIgwAttachement": {
        "Properties": {
            "InternetGatewayId": {"Ref": "TestIgw"},
            "VpcId": {"Ref": "VpcTest"},
        },
        "Type": "AWS::EC2::VPCGatewayAttachment",
    },
    "TestIgwRoute": {
        "Properties": {
            "RouteTableId": {"Ref": "TestIgwRouteTable"},
            "DestinationCidrBlock": "0.0.0.0/0",
            "GatewayId": {"Ref": "TestIgw"},
        },
        "Type": "AWS::EC2::Route",
    },
}


@pytest.mark.parametrize("route_table_provided", [False, True])
def test_internet_gateway(stack: Stack, route_table_provided: bool) -> None:
    """Test InternetGateway construct."""
    vpc = ec2.VPC(
        name_to_id("vpc-test"),
        CidrBlock="10.0.0.0/16",
        EnableDnsHostnames="true",
        EnableDnsSupport="true",
    )

    subnets = [
        ec2.Subnet(
            name_to_id(f"{zone}-subnet"),
            CidrBlock="10.0.0.0/20",
            AvailabilityZone="eu-west-1b",
            VpcId=Ref(vpc),
            MapPublicIpOnLaunch="true",
        )
        for zone, ip in zip(  # noqa: B905
            ["eu-west-1a", "eu-west-1b"], ["10.0.0.0/20", "10.0.16.0/20"]
        )
    ]

    if route_table_provided:
        route_table = ec2.RouteTable(name_to_id("TestIgwRouteTable"), VpcId=Ref(vpc))
    else:
        route_table = None

    igw = InternetGateway(
        name_prefix="test", vpc=vpc, subnets=subnets, route_table=route_table
    )

    for el in (vpc, *subnets, igw):
        stack.add(el)

    template = dict(EXPECTED_TEMPLATE)

    if not route_table_provided:
        template["TestIgwRouteTable"] = {
            "Properties": {"VpcId": {"Ref": "VpcTest"}},
            "Type": "AWS::EC2::RouteTable",
        }
        template["Test0"] = {
            "Properties": {
                "RouteTableId": {"Ref": "TestIgwRouteTable"},
                "SubnetId": {"Ref": "EuWest1aSubnet"},
            },
            "Type": "AWS::EC2::SubnetRouteTableAssociation",
        }
        template["Test1"] = {
            "Properties": {
                "RouteTableId": {"Ref": "TestIgwRouteTable"},
                "SubnetId": {"Ref": "EuWest1bSubnet"},
            },
            "Type": "AWS::EC2::SubnetRouteTableAssociation",
        }

    assert stack.export()["Resources"] == template



from __future__ import annotations
import pytest
import argparse
from unittest import mock
from textwrap import dedent
from botocore.credentials import ReadOnlyCredentials
from e3.aws import assume_profile_main
from typing import TYPE_CHECKING

if TYPE_CHECKING:
    from typing import Any


# Output expected when running without --json
EXPECTED_DEFAULT_OUTPUT = dedent(
    """\
    export AWS_ACCESS_KEY_ID=access_key
    export AWS_SECRET_ACCESS_KEY=secret_key
    export AWS_SESSION_TOKEN=token"""
)


# Output expected when running with --json
EXPECTED_JSON_OUTPUT = (
    '{"AccessKeyId": "access_key", "SecretAccessKey": '
    '"secret_key", "SessionToken": "token"}'
)


class MockSession:
    def __init__(self, *args: list[Any], **kwargs: dict[str, Any]) -> None:
        """Mock the boto3.Session class."""
        pass

    def get_credentials(self) -> "MockSession":
        """Return this class for get_frozen_credentials."""
        return self

    def get_frozen_credentials(self) -> ReadOnlyCredentials:
        """Return fake credentials."""
        return ReadOnlyCredentials("access_key", "secret_key", "token")


@pytest.mark.parametrize(
    "json,expected_output",
    [(False, EXPECTED_DEFAULT_OUTPUT), (True, EXPECTED_JSON_OUTPUT)],
)
def test_assume_profile_main_json(json: bool, expected_output: str, capfd):
    """Test the credentials returned by assume_profile_main."""
    with (
        mock.patch(
            "argparse.ArgumentParser.parse_args",
            return_value=argparse.Namespace(json=json, profile="foo"),
        ),
        mock.patch(
            "boto3.Session",
            new=MockSession,
        ),
    ):
        assume_profile_main()

    stdout = capfd.readouterr().out
    assert stdout.strip() == expected_output



from __future__ import annotations
from typing import TYPE_CHECKING
from e3.aws.cfn import AWSType, GetAtt, Resource
from e3.aws.cfn.ec2 import VPC

if TYPE_CHECKING:
    from typing import Any


class GroupSecurityRule:
    """Security rule for EC2 Security groups."""

    RULE_TYPE: str | None = None
    PROTOCOLS: dict[str | int, dict[str, int | str]] = {
        "ssh": {"from": 22, "to": 22, "protocol": "tcp"},
        "smtps": {"from": 465, "to": 465, "protocol": "tcp"},
        "https": {"from": 443, "to": 443, "protocol": "tcp"},
        "http": {"from": 80, "to": 80, "protocol": "tcp"},
        "pip": {"from": 3128, "to": 3128, "protocol": "tcp"},
        "alltcp": {"from": 1, "to": 65535, "protocol": "tcp"},
        "ntp": {"from": 123, "to": 123, "protocol": "udp"},
        "postgresql": {"from": 5432, "to": 5432, "protocol": "tcp"},
    }

    def __init__(
        self,
        protocol: str | int,
        target: Any,
        from_port: int | None = None,
        to_port: int | None = None,
        description: str | None = None,
    ):
        """Initialize a security rule.

        :param protocol: either an ip protocol name or int (see AWS
            documentation) or a protocol defined in PROTOCOLS. For the
            later port range are set automatically
        :param target: correspond to destination (Egress rules) or source
            (Ingress rule). The nature of the object is determined by the
            class used. See documenation for each subclass of
            GroupSecurityRule.
        :param from_port: optional starting port
        :param to_port: optional ending port
        :param description: optional description
        """
        self.target = target
        if protocol in self.PROTOCOLS:
            # Handle common protocols
            p = self.PROTOCOLS[protocol]
            self.ip_protocol = p["protocol"]
            self.from_port = p.get("from", None)
            self.to_port = p.get("to", None)
            self.description = p.get("description", None)
        else:
            self.ip_protocol = protocol
            self.from_port = None
            self.to_port = None
            self.description = None

        if from_port is not None:
            self.from_port = from_port

        if to_port is None:
            if self.to_port is None:
                self.to_port = self.from_port
        else:
            self.to_port = to_port

        if description is not None:
            self.description = description

    @property
    def properties(self):
        result = {self.RULE_TYPE: self.target, "IpProtocol": self.ip_protocol}
        if self.from_port is not None:
            result["FromPort"] = self.from_port
        if self.to_port is not None:
            result["ToPort"] = self.to_port
        if self.description is not None:
            result["Description"] = self.description
        return result


class EgressRule(GroupSecurityRule):
    pass


class IngressRule(GroupSecurityRule):
    pass


class Ipv4EgressRule(EgressRule):
    RULE_TYPE: str = "CidrIp"


class PrefixListEgressRule(EgressRule):
    RULE_TYPE: str = "DestinationPrefixListId"


class Ipv4IngressRule(IngressRule):
    RULE_TYPE: str = "CidrIp"


class SecurityGroup(Resource):
    """EC2 Security group resource."""

    def __init__(self, name, vpc, rules=None, description=None):
        """Initialize a security group.

        :param name: logical name in the stack
        :type name: str
        :param vpc: a VPC to which the rule is attached
        :type vpc: e3.aws.cfn.ec2.VPC
        :param rules: a list of rules to apply (both egress and ingress)
        :type rules: None | list[e3.aws.cfn.ec2.security.GroupSecurityRule]
        :param description: an optional description
        :type description: str | None
        """
        super().__init__(name, kind=AWSType.EC2_SECURITY_GROUP)
        assert isinstance(vpc, VPC)
        self.vpc = vpc
        self.description = description
        self.egress = []
        self.ingress = []
        if rules is not None:
            for rule in rules:
                self.add_rule(rule)

    def add_rule(self, rule: GroupSecurityRule) -> None:
        """Add a rule to the security group.

        :param rule: the rule to add
        """
        if isinstance(rule, IngressRule):
            self.ingress.append(rule)
        elif isinstance(rule, EgressRule):
            self.egress.append(rule)
        else:
            raise AssertionError("a security group rule is expected")

    @property
    def group_id(self):
        """Return SecurityGroup GroupId."""
        return GetAtt(self.name, "GroupId")

    @property
    def properties(self):
        result = {"VpcId": self.vpc.ref}

        if self.egress:
            result["SecurityGroupEgress"] = [r.properties for r in self.egress]
        if self.ingress:
            result["SecurityGroupIngress"] = [r.properties for r in self.ingress]
        if self.description is not None:
            result["GroupDescription"] = self.description
        return result
